#!/usr/bin/env python
# encoding: utf-8
"""
_TaskChain_

Created by Dave Evans on 2011-06-15.
Copyright (c) 2011 Fermilab. All rights reserved.

Provide support for building arbitrary chains of WMTasks based on a nested dictionary structure
starting with either a generation (create new MC events) or processing (use an existing input dataset) step, followed
by a chain of dependent WMTasks that process the subsequent output.

The request is formed as a dictionary where some global parameters are provided as normal, but the
processing tasks are specified as sub dictionaries.

The top level dict should contain the parameter TaskChain and the value is the number of processing tasks to be run.
For each count in the chain, a dictionary entry named Task1...N should be made with a value being another dictionary.

Any parameters in the Main request will be used throughout the different task unless they are overriden, exceptions are
CouchDB parameters the main request parameters are:

{
    "CMSSWVersion": "CMSSW_3_5_8",                    CMSSW Version
    "ScramArch": "slc5_ia32_gcc434",                  Scram Arch
    "Requestor": "sfoulkes@fnal.gov",                 Person responsible
    "GlobalTag": "GR10_P_v4::All",                    Global Tag
    "ConfigCacheUrl": https://cmsweb-testbed.cern.ch/couchdb URL of CouchDB containing ConfigCache docs for all tasks
    "CouchDBName": "config_cache",                    Name of Couch Database containing config cache (Used for all sub-tasks)
    "TaskChain" : 4,                                  Define number of tasks in chain.
}


Task1 will be either a generation or processing task:

Example initial generation task:

"Task1" :{
    "TaskName"           : "GenSim",                 Task Name
    "ConfigCacheID"      : generatorDoc,             Generator Config id
    "SplittingAlgorithm" : "EventBased",             Splitting Algorithm
    "SplittingArguments" : {"events_per_job" : 250}, Size of jobs in terms of splitting algorithm
    "RequestNumEvents"   : 10000,                    Total number of events to generate
    "Seeding"            : "AutomaticSeeding",       Random seeding method
    "PrimaryDataset"     : "RelValTTBar",            Primary Dataset to be created
    "ScramArch"          : "slc5_amd64_gcc462",      Particular scramArch for this task
    "CMSSWVersion"       : "CMSSW_5_3_5",            Particular CMSSW version for this task
},

Example initial processing task

"Task1" :{
     "TaskName"           : "DigiHLT",                                 Task Name
     "ConfigCacheID"      : someHash,                                  Processing Config id
     "InputDataset"       : "/MinimumBias/Commissioning10-v4/GEN-SIM", Input Dataset to be processed
     "SplittingAlgorithm" : "FileBased",                               Splitting Algorithm
     "SplittingArguments" : {"files_per_job" : 1},                     Size of jobs in terms of splitting algorithm
     "MCPileup"           : "/MinBias/Summer12-v1/GEN-SIM-DIGI-RECO",  Pileup MC dataset for the task
     "DataPileup"         : "/MinimumBias/Run2011A-v1/RAW"             Pileup data dataset for the task
     "GlobalTag"          : "GR_P_V42::All"                            Global tag for  this task
     "KeepOutput"         : False                                      Indicates if the output data from this dataset should be kept in merged area
 },

 All subsequent Task entries will process the output of one of the preceeding steps, the primary dataset can be changed from the input.

 Example:

 "Task2" : {
     "TaskName"              : "Reco",                        Task Name
     "InputTask"             : "DigiHLT",                     Input Task Name (Task Name field of a previous Task entry)
     "InputFromOutputModule" : "writeRAWDIGI",                OutputModule name in the input task that will provide files to process
     "ConfigCacheID"         : "17612875182763812763812",     Processing Config id
     "SplittingAlgorithm"    : "FileBased",                   Splitting Algorithm
     "SplittingArguments"    : {"files_per_job" : 1 },        Size of jobs in terms of splitting algorithm
     "DataPileup"            : "/MinimumBias/Run2011A-v1/RAW" Pileup data dataset for the task
 },

 "Task3" : {
     "TaskName"              : "ALCAReco",             Task Name
     "InputTask"             : "Reco",                 Input Task Name (Task Name field of a previous Task entry)
     "InputFromOutputModule" : "writeALCA",            OutputModule name in the input task that will provide files to process
     "ConfigCacheID"         : "12871372323918187281", Processing Config id
     "SplittingAlgorithm"    : "FileBased",            Splitting Algorithm
     "SplittingArguments"    : {"files_per_job" : 1 }, Size of jobs in terms of splitting algorithm
 },
"""
from __future__ import division

import json
from builtins import range, object
from future.utils import viewitems

from Utils.Utilities import makeList, strToBool
from WMCore.Lexicon import primdataset, taskStepName, gpuParameters
from WMCore.WMSpec.StdSpecs.StdBase import StdBase
from WMCore.WMSpec.WMWorkloadTools import (validateArgumentsCreate, parsePileupConfig,
                                           checkMemCore, checkEventStreams, checkTimePerEvent)
from WMCore.WMSpec.WMSpecErrors import WMSpecFactoryException

#
# simple utils for data mining the request dictionary
#
parentTaskModule = lambda args: args.get("InputFromOutputModule", None)


class ParameterStorage(object):
    """
    _ParameterStorage_

    Decorator class which storages global parameters,
    sets them to local values before executing the passed function
    and restores them afterwards. This is only suited to decorate the
    setupTask and setupGeneratorTask in TaskChainWorkloadFactory.
    """

    def __init__(self, func):
        """
        __init__

        Stores the function and valid parameters to save/restore.

        The validParameters dictionary contains a mapping from the name of the attribute
        in StdBase to the argument key in the task dictionaries
        """
        self.func = func
        self.validParameters = {'globalTag': 'GlobalTag',
                                'frameworkVersion': 'CMSSWVersion',
                                'scramArch': 'ScramArch',
                                'processingVersion': 'ProcessingVersion',
                                'processingString': 'ProcessingString',
                                'acquisitionEra': 'AcquisitionEra',
                                'timePerEvent': 'TimePerEvent',
                                'sizePerEvent': 'SizePerEvent',
                                'memory': 'Memory'
                               }
        return

    def __get__(self, instance, owner):
        """
        __get__

        Get method for the class, store the calling instance for latter use
        """
        self.obj = instance
        self.cls = owner
        return self.__call__

    def __call__(self, task, taskConf):
        """
        __call__

        Store the global parameters, alters the parameters
        using the taskConf argument. Executes the stored
        method, then restores the parameters and resets the local instance.
        """
        self.storeParameters()
        self.alterParameters(taskConf)
        self.func(self.obj, task, taskConf)
        self.restoreParameters()
        self.resetParameters()
        return

    def storeParameters(self):
        """
        _storeParameters_

        Store the original parameters in the decorator
        """
        for param in self.validParameters:
            globalValue = getattr(self.obj, param, None)
            setattr(self, param, globalValue)
        return

    def alterParameters(self, taskConf):
        """
        _alterParameters_

        Alter the parameters with the specific task configuration
        """
        for param in self.validParameters:
            # if task arg is None or 0 or "", then reuse the global one
            taskValue = getattr(self, param)
            if taskConf.get(self.validParameters[param]):
                taskValue = taskConf[self.validParameters[param]]
            setattr(self.obj, param, taskValue)
        return

    def restoreParameters(self):
        """
        _restoreParameters_

        Restore the parameters to the global values
        """
        for param in self.validParameters:
            globalValue = getattr(self, param)
            setattr(self.obj, param, globalValue)
        return

    def resetParameters(self):
        """
        _resetParameters_

        Reset parameters to None
        """
        for param in self.validParameters:
            setattr(self, param, None)
        return


class TaskChainWorkloadFactory(StdBase):
    def __init__(self):
        StdBase.__init__(self)
        self.eventsPerJob = None
        self.eventsPerLumi = None
        self.mergeMapping = {}
        self.taskMapping = {}
        self.taskOutputMapping = {}

    def __call__(self, workloadName, arguments):
        """
        _call_

        Create a ReReco workload with the given parameters.
        """
        StdBase.__call__(self, workloadName, arguments)
        self.workload = self.createWorkload()

        # Detect blow-up factor from first task in chain.
        blowupFactor = 1
        if (self.taskChain > 1) and 'TimePerEvent' in arguments["Task1"]:
            origTpe = arguments["Task1"]['TimePerEvent']
            if origTpe <= 0:
                origTpe = 1.0
            sumTpe = 0
            tpeCount = 0
            for i in range(1, self.taskChain + 1):
                if 'TimePerEvent' in arguments["Task%d" % i]:
                    sumTpe += arguments["Task%d" % i]['TimePerEvent']
                    tpeCount += 1
            if tpeCount > 0:
                blowupFactor = sumTpe / origTpe

        for i in range(1, self.taskChain + 1):

            originalTaskConf = arguments["Task%d" % i]
            taskConf = {}
            # Make a shallow copy of the taskConf
            for k, v in viewitems(originalTaskConf):
                taskConf[k] = v
            parent = taskConf.get("InputTask", None)

            self.modifyTaskConfiguration(taskConf, i == 1, i == 1 and 'InputDataset' not in taskConf)

            # Set task-specific global parameters
            self.blockBlacklist = taskConf["BlockBlacklist"]
            self.blockWhitelist = taskConf["BlockWhitelist"]
            self.runBlacklist = taskConf["RunBlacklist"]
            self.runWhitelist = taskConf["RunWhitelist"]

            parentTask = None
            if parent in self.mergeMapping:
                parentTask = self.mergeMapping[parent][parentTaskModule(taskConf)]

            task = self.makeTask(taskConf, parentTask)

            if i == 1:
                # First task will either be generator or processing
                startPolicy = self.decideWorkQueueStartPolicy(arguments)
                self.workload.setWorkQueueSplitPolicy(startPolicy, taskConf['SplittingAlgo'],
                                                      taskConf['SplittingArguments'],
                                                      blowupFactor=blowupFactor,
                                                      OpenRunningTimeout=self.openRunningTimeout)
                if startPolicy == "MonteCarlo":
                    # generate mc events
                    self.workload.setDashboardActivity("production")
                    self.workload.setEndPolicy("SingleShot")
                    self.setupGeneratorTask(task, taskConf)
                else:
                    # process an existing dataset
                    self.workload.setDashboardActivity("processing")
                    self.setupTask(task, taskConf)
            else:
                # all subsequent tasks have to be processing tasks
                self.setupTask(task, taskConf)
            self.taskMapping[task.name()] = taskConf

        # now that all tasks have been created, create the parent x output dataset map
        self.createTaskParentageMapping(arguments)
        self.workload.setTaskParentageMapping(self.taskOutputMapping)

        self.workload.ignoreOutputModules(self.ignoredOutputModules)
        # and push the parentage map to the reqmgr2 workload cache doc
        arguments['ChainParentageMap'] = self.workload.getChainParentageSimpleMapping()

        # Feed values back to save in couch
        if self.eventsPerJob:
            arguments['Task1']['EventsPerJob'] = self.eventsPerJob
        if self.eventsPerLumi:
            arguments['Task1']['EventsPerLumi'] = self.eventsPerLumi
        return self.workload

    def createTaskParentageMapping(self, origArgs):
        """
        Create a dict struct with a mapping of task name to parent dataset
        and output datasets.
        :param taskO: a WMTask object for the production/processing task
        :param origArgs: arguments provided by the user + default spec args
        :param firstTask: if Task1, then call it with firstTask=True
        :return: update a dictionary in place which will be later set as a
        WMWorkload property
        """
        for i in range(1, self.taskChain + 1):
            taskNumber = "Task%d" % i
            taskName = origArgs[taskNumber]['TaskName']
            taskO = self.workload.getTaskByName(taskName)
            self.taskOutputMapping.setdefault(taskName, {})

            # create a few more key/value pairs to help find the parent. Remove it later
            self.taskOutputMapping[taskName] = {'TaskNumber': taskNumber,
                                                'ParentTaskNumber': None,
                                                'ParentTaskName': None,
                                                'ParentDataset': None,
                                                'OutputDatasetMap': {}}

            if taskNumber == 'Task1':
                self.taskOutputMapping[taskName]['ParentDataset'] = origArgs[taskNumber].get('InputDataset')
            else:
                parentTaskName = origArgs[taskNumber]["InputTask"]
                parentOutputModName = origArgs[taskNumber]["InputFromOutputModule"]
                parentTaskNumber = self.taskOutputMapping[parentTaskName]['TaskNumber']

                self.taskOutputMapping[taskName]['ParentTaskName'] = parentTaskName
                self.taskOutputMapping[taskName]['ParentTaskNumber'] = parentTaskNumber

                parentDset = self.findParentDataset(origArgs, parentTaskNumber, parentTaskName, parentOutputModName)
                self.taskOutputMapping[taskName]['ParentDataset'] = parentDset

            # set the OutputDatasetMap or empty if KeepOutput is False
            if not origArgs[taskNumber].get("KeepOutput", True):
                continue

            transientMods = origArgs[taskNumber].get("TransientOutputModules", [])
            for outInfo in taskO.listOutputDatasetsAndModules():
                outModName = outInfo['outputModule']
                if outModName not in transientMods:
                    self.taskOutputMapping[taskName]['OutputDatasetMap'][outModName] = outInfo['outputDataset']

    def findParentDataset(self, origArgs, taskNumber, taskName, outModName):
        """
        _findParentDataset_
        Given the parent task name and output module name, finds the parent dataset
        :param origArgs: request arguments
        :param taskNumber: step number of the parent step
        :param taskName: step name of the parent step
        :param outModName: output module name of the parent step
        :return: the parent dataset name (str), otherwise None
        """
        # mind there might be transient output modules too...
        if outModName in self.taskOutputMapping[taskName]['OutputDatasetMap']:
            return self.taskOutputMapping[taskName]['OutputDatasetMap'][outModName]
        else:
            # then fetch grand-parent data
            parentTaskNumber = self.taskOutputMapping[taskName]['ParentTaskNumber']
            parentTaskName = self.taskOutputMapping[taskName]['ParentTaskName']
            if parentTaskNumber:
                parentOutputModName = origArgs[taskNumber]["InputFromOutputModule"]
                return self.findParentDataset(origArgs, parentTaskNumber, parentTaskName, parentOutputModName)
            else:
                # this is Task1
                return self.taskOutputMapping[taskName]['ParentDataset']

    def makeTask(self, taskConf, parentTask=None):
        """
        _makeTask_

        create new Task and populate it with basic required parameters from the
        taskConfig provided, if parentTask is None, the task will be created in
        the workload, else the task will be created as a child of the parent Task

        """
        if parentTask is None:
            task = self.workload.newTask(taskConf['TaskName'])
        else:
            task = parentTask.addTask(taskConf['TaskName'])
        return task

    def _updateCommonParams(self, task, taskConf):
        # sets the prepID  all the properties need to be set by
        # self.workload.setTaskPropertiesFromWorkload manually for the task
        task.setPrepID(taskConf.get("PrepID", self.workload.getPrepID()))
        task.setAcquisitionEra(taskConf.get("AcquisitionEra", self.workload.getAcquisitionEra()))
        task.setProcessingString(taskConf.get("ProcessingString", self.workload.getProcessingString()))
        task.setProcessingVersion(taskConf.get("ProcessingVersion", self.workload.getProcessingVersion()))
        lumiMask = taskConf.get("LumiList", self.workload.getLumiList())
        if lumiMask:
            task.setLumiMask(lumiMask)

        if taskConf["PileupConfig"]:
            self.setupPileup(task, taskConf['PileupConfig'])

    @ParameterStorage
    def setupGeneratorTask(self, task, taskConf):
        """
        _setupGeneratorTask_

        Set up an initial generation task
        """
        cmsswStepType = "CMSSW"
        configCacheID = taskConf['ConfigCacheID']
        splitAlgorithm = taskConf["SplittingAlgo"]
        splitArguments = taskConf["SplittingArguments"]
        keepOutput = taskConf["KeepOutput"]
        transientModules = taskConf["TransientOutputModules"]
        forceUnmerged = (not keepOutput) or (len(transientModules) > 0)
        cmsswVersion = taskConf.get('CMSSWVersion', self.frameworkVersion)
        scramArch = taskConf.get('ScramArch', self.scramArch)
        self.inputPrimaryDataset = taskConf['PrimaryDataset']
        outputMods = self.setupProcessingTask(task, "Production", couchDBName=self.couchDBName,
                                              configDoc=configCacheID, configCacheUrl=self.configCacheUrl,
                                              splitAlgo=splitAlgorithm, splitArgs=splitArguments,
                                              stepType=cmsswStepType, seeding=taskConf['Seeding'],
                                              totalEvents=taskConf['RequestNumEvents'],
                                              forceUnmerged=forceUnmerged,
                                              timePerEvent=taskConf.get('TimePerEvent', None),
                                              sizePerEvent=taskConf.get('SizePerEvent', None),
                                              memoryReq=taskConf.get('Memory', None),
                                              cmsswVersion=cmsswVersion,
                                              scramArch=scramArch,
                                              taskConf=taskConf)

        self.addLogCollectTask(task, 'LogCollectFor%s' % task.name(), cmsswVersion=cmsswVersion, scramArch=scramArch)

        # Do the output module merged/unmerged association
        self.setUpMergeTasks(task, outputMods, splitAlgorithm, keepOutput, transientModules,
                             cmsswVersion=cmsswVersion, scramArch=scramArch)

        # this need to be called after setpuProcessingTask since it will overwrite some values
        self._updateCommonParams(task, taskConf)

        return

    @ParameterStorage
    def setupTask(self, task, taskConf):
        """
        _setupTask_

        Build the task using the setupProcessingTask from StdBase
        and set the parents appropriately to handle a processing task
        """

        cmsswStepType = "CMSSW"
        configCacheID = taskConf["ConfigCacheID"]
        splitAlgorithm = taskConf["SplittingAlgo"]
        splitArguments = taskConf["SplittingArguments"]
        keepOutput = taskConf["KeepOutput"]
        transientModules = taskConf["TransientOutputModules"]
        forceUnmerged = (not keepOutput) or (len(transientModules) > 0)
        cmsswVersion = taskConf.get('CMSSWVersion', self.frameworkVersion)
        scramArch = taskConf.get('ScramArch', self.scramArch)

        # in case the initial task is a processing task, we have an input dataset, otherwise
        # we look up the parent task and step
        inputDataset = taskConf.get("InputDataset")
        if inputDataset is not None:
            self.inputDataset = inputDataset
            (self.inputPrimaryDataset, self.inputProcessedDataset,
             self.inputDataTier) = self.inputDataset[1:].split("/")
            inpStep = None
            inpMod = None
        else:
            self.inputDataset = None
            inputTask = taskConf["InputTask"]
            inputTaskConf = self.taskMapping[inputTask]
            parentTaskForMod = self.mergeMapping[inputTask][taskConf['InputFromOutputModule']]
            inpStep = parentTaskForMod.getStep("cmsRun1")
            if not inputTaskConf["KeepOutput"] or len(inputTaskConf["TransientOutputModules"]) > 0:
                inpMod = taskConf["InputFromOutputModule"]
                # Check if the splitting has to be changed
                if inputTaskConf["SplittingAlgo"] == 'EventBased' \
                        and (inputTaskConf.get("InputDataset") or inputTaskConf.get("InputTask")):
                    splitAlgorithm = 'WMBSMergeBySize'
                    splitArguments = {'max_merge_size': self.maxMergeSize,
                                      'min_merge_size': self.minMergeSize,
                                      'max_merge_events': self.maxMergeEvents,
                                      'max_wait_time': self.maxWaitTime}
            else:
                inpMod = "Merged"

        currentPrimaryDataset = self.inputPrimaryDataset
        if taskConf.get("PrimaryDataset") is not None:
            self.inputPrimaryDataset = taskConf.get("PrimaryDataset")

        outputMods = self.setupProcessingTask(task, "Processing",
                                              inputDataset,
                                              inputStep=inpStep,
                                              inputModule=inpMod,
                                              couchDBName=self.couchDBName,
                                              configCacheUrl=self.configCacheUrl,
                                              configDoc=configCacheID,
                                              splitAlgo=taskConf["SplittingAlgo"],
                                              splitArgs=splitArguments,
                                              stepType=cmsswStepType,
                                              forceUnmerged=forceUnmerged,
                                              timePerEvent=taskConf.get('TimePerEvent', None),
                                              sizePerEvent=taskConf.get('SizePerEvent', None),
                                              memoryReq=taskConf.get("Memory", None),
                                              cmsswVersion=cmsswVersion,
                                              scramArch=scramArch,
                                              taskConf=taskConf)

        self.addLogCollectTask(task, 'LogCollectFor%s' % task.name(), cmsswVersion=cmsswVersion, scramArch=scramArch)
        self.setUpMergeTasks(task, outputMods, splitAlgorithm, keepOutput, transientModules,
                             cmsswVersion=cmsswVersion, scramArch=scramArch)

        self.inputPrimaryDataset = currentPrimaryDataset

        # this need to be called after setpuProcessingTask since it will overwrite some values
        self._updateCommonParams(task, taskConf)

        return

    def setUpMergeTasks(self, parentTask, outputModules, splittingAlgo,
                        keepOutput, transientOutputModules, cmsswVersion=None, scramArch=None):
        """
        _setUpMergeTasks_

        Set up the required merged tasks according to the following parameters:
        - KeepOutput : All output modules not in the transient list are merged.
        - TransientOutputModules : These output modules won't be merged.
        If not merged then only a cleanup task is created.
        """
        modulesToMerge = []
        unmergedModules = list(outputModules)
        if keepOutput:
            unmergedModules = [x for x in outputModules if x in transientOutputModules]
            modulesToMerge = [x for x in outputModules if x not in transientOutputModules]

        procMergeTasks = {}
        for outputModuleName in modulesToMerge:
            mergeTask = self.addMergeTask(parentTask, splittingAlgo,
                                          outputModuleName, cmsswVersion=cmsswVersion, scramArch=scramArch)
            procMergeTasks[str(outputModuleName)] = mergeTask
        self.mergeMapping[parentTask.name()] = procMergeTasks

        procTasks = {}
        for outputModuleName in unmergedModules:
            self.addCleanupTask(parentTask, outputModuleName, dataTier=outputModules[outputModuleName]['dataTier'])
            procTasks[outputModuleName] = parentTask
        self.mergeMapping[parentTask.name()].update(procTasks)

        return

    def modifyTaskConfiguration(self, taskConf,
                                firstTask=False, generator=False):
        """
        _modifyTaskConfiguration_

        Modify the TaskConfiguration according to the specifications
        in getWorkloadCreateArgs and getChainCreateArgs.
        It does type casting and assigns default values if key is not
        present, unless default value is None.
        """
        taskArguments = self.getChainCreateArgs(firstTask, generator)
        for argument in taskArguments:
            if argument not in taskConf and taskArguments[argument]["default"] is not None:
                taskConf[argument] = taskArguments[argument]["default"]
            elif argument in taskConf:
                taskConf[argument] = taskArguments[argument]["type"](taskConf[argument])

        if generator:
            taskConf["SplittingAlgo"] = "EventBased"
            # Adjust totalEvents according to the filter efficiency
            taskConf["RequestNumEvents"] = int(taskConf.get("RequestNumEvents", 0) / \
                                               taskConf.get("FilterEfficiency"))
            taskConf["SizePerEvent"] = taskConf.get("SizePerEvent", self.sizePerEvent) * \
                                       taskConf.get("FilterEfficiency")

        taskConf["SplittingArguments"] = {}
        if taskConf["SplittingAlgo"] in ["EventBased", "EventAwareLumiBased"]:
            taskConf["EventsPerJob"], taskConf["EventsPerLumi"] = StdBase.calcEvtsPerJobLumi(taskConf.get("EventsPerJob"),
                                                                                             taskConf.get("EventsPerLumi"),
                                                                                             taskConf.get("TimePerEvent",
                                                                                                          self.timePerEvent),
                                                                                             taskConf.get("RequestNumEvents"))
            if firstTask:
                self.eventsPerJob = taskConf["EventsPerJob"]
                self.eventsPerLumi = taskConf["EventsPerLumi"]
            taskConf["SplittingArguments"]["events_per_job"] = taskConf["EventsPerJob"]
            if taskConf["SplittingAlgo"] == "EventBased":
                taskConf["SplittingArguments"]["events_per_lumi"] = taskConf["EventsPerLumi"]
            else:
                taskConf["SplittingArguments"]["job_time_limit"] = 48 * 3600  # 2 days
            taskConf["SplittingArguments"]["lheInputFiles"] = taskConf["LheInputFiles"]
        elif taskConf["SplittingAlgo"] == "LumiBased":
            taskConf["SplittingArguments"]["lumis_per_job"] = taskConf["LumisPerJob"]
        elif taskConf["SplittingAlgo"] == "FileBased":
            taskConf["SplittingArguments"]["files_per_job"] = taskConf["FilesPerJob"]

        taskConf["SplittingArguments"].setdefault("include_parents", taskConf['IncludeParents'])

        taskConf["PileupConfig"] = parsePileupConfig(taskConf.get("MCPileup"),
                                                     taskConf.get("DataPileup"))
        # Adjust the pileup splitting
        taskConf["SplittingArguments"].setdefault("deterministicPileup", taskConf['DeterministicPileup'])

        return

    @staticmethod
    def getWorkloadCreateArgs():
        baseArgs = StdBase.getWorkloadCreateArgs()
        specArgs = {"RequestType": {"default": "TaskChain", "optional": False},
                    "Task1": {"default": {}, "optional": False, "type": dict},
                    # ConfigCacheID is not used in the main dict for TaskChain
                    "ConfigCacheID": {"optional": True, "null": True},
                    "IgnoredOutputModules": {"default": [], "type": makeList, "null": False},
                    "TaskChain": {"default": 1, "type": int,
                                  "optional": False, "validate": lambda x: x > 0,
                                  "attr": "taskChain", "null": False},
                    "ChainParentageMap": {"default": {}, "type": dict},
                    "FirstEvent": {"default": 1, "type": int,
                                   "optional": True, "validate": lambda x: x > 0,
                                   "attr": "firstEvent", "null": False},
                    "FirstLumi": {"default": 1, "type": int,
                                  "optional": True, "validate": lambda x: x > 0,
                                  "attr": "firstLumi", "null": False},
                    ### Override StdBase parameter definition
                    "TimePerEvent": {"default": 12.0, "type": float, "validate": checkTimePerEvent},
                    "Memory": {"default": 2300.0, "type": float, "validate": checkMemCore},
                    "Multicore": {"default": 1, "type": int, "validate": checkMemCore},
                    "EventStreams": {"type": int, "null": True, "default": 0, "validate": checkEventStreams},
                    # no need for workload-level defaults, if task-level default is provided
                    "RequiresGPU": {"default": None, "null": True,
                                    "validate": lambda x: x in ("forbidden", "optional", "required")},
                    "GPUParams": {"default": json.dumps(None), "validate": gpuParameters},
                    }
        baseArgs.update(specArgs)
        StdBase.setDefaultArgumentsProperty(baseArgs)
        return baseArgs

    @staticmethod
    def getChainCreateArgs(firstTask=False, generator=False):
        """
        _getChainCreateArgs_

        Each task dictionary specifies its own set of arguments
        that need to be validated as well, most of them are already
        defined in StdBase.getWorkloadCreateArgs and those do not appear here
        since they are all optional. Here only new arguments are listed.
        """
        baseArgs = StdBase.getChainCreateArgs(firstTask, generator)
        arguments = {
            "TaskName": {"optional": False, "null": False, 'validate': taskStepName},
            "InputTask": {"default": None, "optional": firstTask, "null": False},
            "TransientOutputModules": {"default": [], "type": makeList, "optional": True, "null": False},
            "DeterministicPileup": {"default": False, "type": strToBool, "optional": True, "null": False},
            "GlobalTag": {"type": str, "optional": True},
            "TimePerEvent": {"type": float, "optional": True, "validate": lambda x: x > 0},
            "SizePerEvent": {"type": float, "optional": True, "validate": lambda x: x > 0},
            'PrimaryDataset': {'default': None, 'optional': not generator, 'validate': primdataset,
                               'null': False},
                    }
        baseArgs.update(arguments)
        StdBase.setDefaultArgumentsProperty(baseArgs)
        return baseArgs

    @staticmethod
    def getWorkloadAssignArgs():
        baseArgs = StdBase.getWorkloadAssignArgs()
        specArgs = {
            "ChainParentageMap": {"default": {}, "type": dict},
            ### Override StdBase assignment parameter definition
            "Memory": {"type": float, "validate": checkMemCore},
            "Multicore": {"type": int, "validate": checkMemCore},
            "EventStreams": {"type": int, "validate": checkEventStreams},
        }
        baseArgs.update(specArgs)
        StdBase.setDefaultArgumentsProperty(baseArgs)
        return baseArgs

    def validateSchema(self, schema):
        """
        _validateSchema_

        Go over each task and make sure it matches validation
        parameters derived from Dave's requirements. They are:
         * cannot have more than 10 tasks
         * output from the last task *must* be saved
         * task must be described as a python dictionary type
         * transient output modules must be an input for further task(s)
         * and the usual Task arguments validation, as defined in the spec
        """
        numTasks = schema['TaskChain']
        if numTasks > 10:
            msg = "Workflow exceeds the maximum allowed number of tasks. "
            msg += "Limited to up to 10 tasks, found %s tasks." % numTasks
            self.raiseValidationException(msg)

        lastTask = "Task%s" % schema['TaskChain']
        if not strToBool(schema[lastTask].get('KeepOutput', True)):
            msg = "Dropping the output (KeepOutput=False) of the last task is prohibited.\n"
            msg += "You probably want to remove that task completely and try again."
            self.raiseValidationException(msg=msg)

        transientMapping = {}
        for i in range(1, numTasks + 1):
            taskNumber = "Task%s" % i
            if taskNumber not in schema:
                msg = "Task '%s' not present in the request schema." % taskNumber
                self.raiseValidationException(msg=msg)

            task = schema[taskNumber]
            # We can't handle non-dictionary tasks
            if not isinstance(task, dict):
                msg = "Task '%s' not defined as a dictionary. " % taskNumber
                msg += "It could be an indicator of JSON error.\n"
                self.raiseValidationException(msg=msg)

            # Generic task parameter validation
            self.validateTask(task, self.getChainCreateArgs(i == 1, i == 1 and 'InputDataset' not in task))

            # Validate the existence of the configCache
            if task["ConfigCacheID"]:
                self.validateConfigCacheExists(configID=task['ConfigCacheID'],
                                               configCacheUrl=schema["ConfigCacheUrl"],
                                               couchDBName=schema["CouchDBName"],
                                               getOutputModules=False)

            # Validate the chaining of transient output modules, need to make a copy of the lists
            transientMapping[task['TaskName']] = [x for x in task.get('TransientOutputModules', [])]

            if i > 1:
                inputTransientModules = transientMapping[task['InputTask']]
                if task['InputFromOutputModule'] in inputTransientModules:
                    inputTransientModules.remove(task['InputFromOutputModule'])

        try:
            StdBase.validateGPUSettings(schema)
        except Exception as ex:
            self.raiseValidationException(str(ex))

        for task in transientMapping:
            if transientMapping[task]:
                msg = "A transient module is not processed by a subsequent task.\n"
                msg += "This is a malformed task chain workload"
                self.raiseValidationException(msg)

    def validateTask(self, taskConf, taskArgumentDefinition):
        """
        _validateTask_

        Validate the task information against the given
        argument description
        """
        try:
            validateArgumentsCreate(taskConf, taskArgumentDefinition, checkInputDset=False)
            # Validate GPU-related spec parameters
            StdBase.validateGPUSettings(taskConf)
        except WMSpecFactoryException:
            # just re-raise it to keep the error message clear
            raise
        except Exception as ex:
            self.raiseValidationException(str(ex))

    def decideWorkQueueStartPolicy(self, reqDict):
        """
        Given a request dictionary, decides which WorkQueue start
        policy needs to be used in a given request.
        :param reqDict: a dictionary with the creation request information
        :return: a string with the start policy to be used.
        """
        if not reqDict["Task1"].get("InputDataset"):
            return "MonteCarlo"

        inputDset = reqDict["Task1"]["InputDataset"]
        if inputDset.endswith("/MINIAODSIM"):
            return "Dataset"
        else:
            return "Block"



"""
File       : MSOutputStreamer.py

Description: MSOutputStreamer.py class is a simple wrapper for the Output data
placement service in WMCore MicroServices.
"""

# futures
from __future__ import division, print_function

from builtins import object
import json
# from itertools import islice
# from collections import deque


class MSOutputStreamer(object):
    """
    MSOutputStreamer class
    """
    # NOTE: This module here is just a placeholder for future development
    #       It is placed here so we can make the basic construction of the MSOutput module

    def __init__(self, requestRecords=None, bufferFile=None, stride=1, logger=None):
        """
        Creates a basic streamer object.
        For the purpose of testing this two modes of streaming are defined:
           - From a dictionary buffer
           - From a file buffer
        :requestRecords: - a dictionary with request records
        :bufferFile:     - a buffer file containing the same information as :requestRecords:
        :stride:         - the stride at which to stream - N records at a time
        """

        self.requestRecords = requestRecords
        self.bufferFile = bufferFile
        self.logger = logger
        self.stride = stride

    def __call__(self, *args, **kwargs):
        """
        The call method from the Streamer class
        """
        return self.streamer()

    def streamer(self):
        """
        The default streamer function
        NOTE:
            If provided bufferFile takes precedence
        TODO:
            For implementing the streaming for stride != 1 we should use:
            itertools.islice(iterable, start, stop[, step]) + collections.deque() 
        """
        try:
            with open(self.bufferFile) as fp:
                result = json.load(fp)
            requests = result['result'][0]
        except Exception as ex:
            msg = "Could not open bufferFile: %s. Exception %s" % (self.bufferFile, str(ex))
            msg += "Falling back to dictionary buffer."
            self.logger.info(msg)
            requests = self.requestRecords

        if requests is None:
            requests = {}

        while requests:
            yield requests.popitem()



#!/usr/bin/env python
"""
_LoadDBSFilesByDAS_

MySQL implementation of LoadDBSFilesByDAS
"""

from WMCore.Database.DBFormatter import DBFormatter

class LoadDBSFilesByDAS(DBFormatter):
    fileInfoSQL = """SELECT dbsbuffer_file.id AS id,
                            dbsbuffer_file.lfn AS lfn,
                            dbsbuffer_file.filesize AS filesize,
                            dbsbuffer_file.events AS events,
                            dbsbuffer_file.status AS status,
                            dbsbuffer_file.block_id AS block,
                            dbsbuffer_algo.app_name AS app_name,
                            dbsbuffer_algo.app_ver AS app_ver,
                            dbsbuffer_algo.app_fam AS app_fam,
                            dbsbuffer_algo.pset_hash AS pset_hash,
                            dbsbuffer_algo.config_content,
                            dbsbuffer_dataset.path AS dataset_path,
                            dbsbuffer_dataset.global_tag AS global_tag,
                            dbsbuffer_dataset.prep_id AS prep_id,
                            dbsbuffer_workflow.name AS workflow,
                            dbsbuffer_workflow.block_close_max_wait_time,
                            dbsbuffer_workflow.block_close_max_files,
                            dbsbuffer_workflow.block_close_max_events,
                            dbsbuffer_workflow.block_close_max_size
                     FROM dbsbuffer_file
                     INNER JOIN dbsbuffer_algo_dataset_assoc ON
                       dbsbuffer_file.dataset_algo = dbsbuffer_algo_dataset_assoc.id
                     INNER JOIN dbsbuffer_algo ON
                       dbsbuffer_algo_dataset_assoc.algo_id = dbsbuffer_algo.id
                     INNER JOIN dbsbuffer_dataset ON
                       dbsbuffer_algo_dataset_assoc.dataset_id = dbsbuffer_dataset.id AND
                       dbsbuffer_dataset.path = :datasetpath
                     INNER JOIN dbsbuffer_workflow ON
                       dbsbuffer_workflow.id = dbsbuffer_file.workflow
                     WHERE dbsbuffer_file.status = 'NOTUPLOADED'
                     AND NOT EXISTS ( SELECT *
                                      FROM dbsbuffer_file_parent
                                      INNER JOIN dbsbuffer_file parent_file ON
                                        parent_file.id = dbsbuffer_file_parent.parent AND
                                        parent_file.status = 'NOTUPLOADED'
                                      WHERE dbsbuffer_file_parent.child = dbsbuffer_file.id )
                     ORDER BY dbsbuffer_file.id
                     """

    getLocationSQL = """SELECT dbsbuffer_location.pnn as location, dbsbuffer_file.id as id
                          FROM dbsbuffer_location
                          INNER JOIN dbsbuffer_file_location dfl ON dfl.location = dbsbuffer_location.id
                          INNER JOIN dbsbuffer_file ON dbsbuffer_file.id = dfl.filename
                          WHERE dbsbuffer_file.id = :fileid"""


    getChecksumSQL = """SELECT cst.type AS cktype, fcs.cksum AS cksum, fcs.fileid AS id FROM
                           dbsbuffer_file_checksums fcs INNER JOIN
                           dbsbuffer_checksum_type cst ON fcs.typeid = cst.id
                           WHERE fcs.fileid = :fileid"""


    getRunLumiSQL = """SELECT flr.run AS run, flr.lumi AS lumi, flr.num_events AS num_events, dbsbuffer_file.id AS id
                          FROM dbsbuffer_file_runlumi_map flr
                          INNER JOIN dbsbuffer_file ON dbsbuffer_file.id = flr.filename
                          WHERE dbsbuffer_file.id = :fileid"""


    getParentLFNSQL = """SELECT dbfa.lfn AS lfn, dbfb.id AS id FROM dbsbuffer_file dbfa
                            INNER JOIN dbsbuffer_file_parent dfp ON dfp.parent = dbfa.id
                            INNER JOIN dbsbuffer_file dbfb ON dfp.child = dbfb.id
                            WHERE dbfb.id = :fileid """






    def formatFileInfo(self, result):
        """
        _formatFileInfo_

        Some databases (Oracle) aren't case sensitive with respect to column
        names so we'll do some formatting so the column names are returned as
        expected.
        """
        resultList = self.formatDict(result)

        for resultDict in resultList:
            resultDict["appName"] = resultDict["app_name"]
            del resultDict["app_name"]

            resultDict["appVer"] = resultDict["app_ver"]
            del resultDict["app_ver"]

            resultDict["appFam"] = resultDict["app_fam"]
            del resultDict["app_fam"]

            resultDict["psetHash"] = resultDict["pset_hash"]
            del resultDict["pset_hash"]

            resultDict["configContent"] = resultDict["config_content"]
            del resultDict["config_content"]

            resultDict["datasetPath"] = resultDict["dataset_path"]
            del resultDict["dataset_path"]

            resultDict["size"] = resultDict["filesize"]
            del resultDict["filesize"]

            resultDict["globalTag"] = resultDict['global_tag']
            del resultDict['global_tag']

        return resultList


    def locInfo(self, result):
        """
        Format the location info so that it matches

        """

        resultList = self.formatDict(result)

        interimDictionary = {}

        for entry in resultList:
            if entry['id'] not in interimDictionary:
                interimDictionary[entry['id']] = set()
            interimDictionary[entry['id']].add(entry['location'])


        finalList = []
        for entry in interimDictionary:
            tmpDict = {'id': entry, 'locations': interimDictionary[entry]}
            finalList.append(tmpDict)


        return finalList



    def ckInfo(self, result):
        """
        Assemble the checksums into the appropriate format.

        """


        resultList = self.formatDict(result)

        interimDictionary = {}

        for entry in resultList:
            if entry['id'] not in interimDictionary:
                interimDictionary[entry['id']] = {}
            interimDictionary[entry['id']][entry['cktype']] = entry['cksum']


        finalList = []
        for entry in interimDictionary:
            tmpDict = {'id': entry, 'checksums': interimDictionary[entry]}
            finalList.append(tmpDict)


        return finalList


    def runInfo(self, result):
        """
        Assemble runLumis into the appropriate format

        """

        resultList = self.formatDict(result)

        interimDictionary = {}

        for entry in resultList:
            if entry['id'] not in interimDictionary:
                interimDictionary[entry['id']] = {}
            if entry['run'] not in interimDictionary[entry['id']]:
                interimDictionary[entry['id']][entry['run']] = []
            interimDictionary[entry['id']][entry['run']].append((entry['lumi'], entry["num_events"]))

        finalList = []
        for entry in interimDictionary:
            tmpDict = {'id': entry, 'runInfo': interimDictionary[entry]}
            finalList.append(tmpDict)


        return finalList


    def parentInfo(self, result):
        """
        Format the parent info so that it makes sense

        """

        resultList = self.formatDict(result)

        interimDictionary = {}

        for entry in resultList:
            if entry['id'] not in interimDictionary:
                interimDictionary[entry['id']] = []
            interimDictionary[entry['id']].append(entry['lfn'])


        finalList = []
        for entry in interimDictionary:
            tmpDict = {'id': entry, 'parentLFNs': interimDictionary[entry]}
            finalList.append(tmpDict)


        return finalList



    def getBinds(self, files):
        binds = []
        files = self.dbi.makelist(files)
        for f in files:
            binds.append({'fileid': f})
        return binds

    def execute(self, datasetpath, conn = None, transaction = False):
        """
        Execute multiple SQL queries to extract all binding information
        Use the first query to get the fileIDs

        """
        result   = self.dbi.processData(self.fileInfoSQL,
                                        { 'datasetpath' : datasetpath },
                                        conn = conn,
                                        transaction = transaction)
        fileInfo = self.formatFileInfo(result)

        fileIDs  = [x['id'] for x in fileInfo]
        binds    = self.getBinds(fileIDs)

        if len(fileInfo) == 0:
            # Then we have no files for this DAS
            return []


        # Do locations
        result   = self.dbi.processData(self.getLocationSQL, binds,
                                        conn = conn,
                                        transaction = transaction)
        locInfo  = self.locInfo(result)
        fullResults = self.merge(fileInfo, locInfo)


        # Do checksums
        result   = self.dbi.processData(self.getChecksumSQL, binds,
                                        conn = conn,
                                        transaction = transaction)

        ckInfo      = self.ckInfo(result)
        fullResults = self.merge(fullResults, ckInfo)


        # Do runLumi
        result      = self.dbi.processData(self.getRunLumiSQL, binds,
                                           conn = conn,
                                           transaction = transaction)
        runInfo  = self.runInfo(result)
        fullResults = self.merge(fullResults, runInfo)



        # Do parents
        result   = self.dbi.processData(self.getParentLFNSQL, binds,
                                        conn = conn,
                                        transaction = transaction)
        parInfo  = self.parentInfo(result)
        fullResults = self.merge(fullResults, parInfo)



        return fullResults


    def merge(self, listA, listB, field = 'id'):
        """
        _merge_

        Merge together two file lists based on the ID field
        """

        for entryA in listA:
            for entryB in listB:
                if entryA[field] == entryB[field]:
                    # Then we've found a match
                    entryA.update(entryB)
                    break


        return listA


    def groupByID(self, inputList, key):
        """
        Group all the entries in a list of dictionaries together by ID


        """

        interimDictionary = {}

        for entry in inputList:
            if entry['id'] not in interimDictionary:
                interimDictionary[entry['id']] = set()
            interimDictionary[entry['id']].add(entry[key])


        finalList = []
        for entry in interimDictionary:
            tmpDict = {'id': entry, key: interimDictionary[entry]}
            finalList.append(tmpDict)


        return finalList



"""
_UpdateLocation_

MySQL implementation of Jobs.UpdateLocation

Created on Apr 17, 2013

@author: dballest
"""

from WMCore.Database.DBFormatter import DBFormatter

class UpdateLocation(DBFormatter):
    """
    _UpdateLocation_

    Update the location of the job in WMBS, must be a valid
    site in wmbs_location.
    """

    sql = """
          UPDATE wmbs_job
          SET location = COALESCE((SELECT id FROM wmbs_location WHERE cms_name = :location), location)
          WHERE id = :jobID
          """

    def execute(self, jobs, conn = None, transaction = False):
        binds = []
        for job in jobs:
            bind = {'jobID' : job['jobid'],
                    'location' : job['location']}
            binds.append(bind)
        self.dbi.processData(self.sql, binds, conn = conn,
                             transaction = transaction)
        return



#!/usr/bin/env python
"""
_Tier1ReRecoWorkload_



"""
from WMCore.WMSpec.WMWorkload import newWorkload


DBSURL = "https://cmsweb-prod.cern.ch/dbs/prod/global/DBSReader"

#  //
# // Set up the basic workload task and step structure
#//
def createWorkload(name="MultiTaskProcessing"):
    workload = newWorkload("MultiTaskReReco")
    workload.setOwner("DMWMTest")
    workload.setStartPolicy('DatasetBlock')
    workload.setEndPolicy('SingleShot')

    #  //
    # // set up the production task
    #//
    rereco = workload.newTask("ReReco1")
    rerecoCmssw = rereco.makeStep("cmsRun1")
    rerecoCmssw.setStepType("CMSSW")
    skimStageOut = rerecoCmssw.addStep("stageOut1")
    skimStageOut.setStepType("StageOut")
    skimLogArch = rerecoCmssw.addStep("logArch1")
    skimLogArch.setStepType("LogArchive")
    rereco.applyTemplates()
    rereco.setSplittingAlgorithm("FileBased", files_per_job = 1)
    rereco.addInputDataset(
        primary = "Cosmics",
        processed = "ComissioningHI-PromptReco-v1",
        tier = "RECO",
        dbsurl = DBSURL)

    #  //
    # // rereco cmssw step
    #//
    #
    # TODO: Anywhere helper.data is accessed means we need a method added to the
    # type based helper class to provide a clear API.
    rerecoCmsswHelper = rerecoCmssw.getTypeHelper()


    rerecoCmsswHelper.cmsswSetup(
        "CMSSW_3_1_2",
        softwareEnvironment = " . /uscmst1/prod/sw/cms/bashrc prod"
        )

    rerecoCmsswHelper.setDataProcessingConfig(
        "cosmics", "promptReco", globalTag = "GLOBAL::BALLS",
        writeTiers = ['RECO'])

    rerecoCmsswHelper.addOutputModule(
        "outputRECO", primaryDataset = "Primary",
        processedDataset = "Processed",
        dataTier = "RECO")

    #Add a stageOut step
    #skimStageOutHelper = skimStageOut.getTypeHelper()
    #skimLogArchHelper  = skimLogArch.getTypeHelper()


    rereco.addGenerator("BasicNaming")

    #  //
    # // set up the production task
    #//
    rereco = workload.newTask("ReReco2")
    rerecoCmssw = rereco.makeStep("cmsRun1")
    rerecoCmssw.setStepType("CMSSW")
    skimStageOut = rerecoCmssw.addStep("stageOut1")
    skimStageOut.setStepType("StageOut")
    skimLogArch = rerecoCmssw.addStep("logArch1")
    skimLogArch.setStepType("LogArchive")
    rereco.applyTemplates()
    rereco.setSplittingAlgorithm("FileBased", files_per_job = 1)
    rereco.setSiteWhitelist(["T2_XX_SiteA", "T2_XX_SiteB", "T2_XX_SiteC"])
    rereco.addInputDataset(
        primary = "Cosmics",
        processed = "ComissioningHI-PromptReco-v1",
        tier = "RECO",
        dbsurl = DBSURL)

    #  //
    # // rereco cmssw step
    #//
    #
    # TODO: Anywhere helper.data is accessed means we need a method added to the
    # type based helper class to provide a clear API.
    rerecoCmsswHelper = rerecoCmssw.getTypeHelper()


    rerecoCmsswHelper.cmsswSetup(
        "CMSSW_3_1_2",
        softwareEnvironment = " . /uscmst1/prod/sw/cms/bashrc prod"
        )

    rerecoCmsswHelper.setDataProcessingConfig(
        "cosmics", "promptReco", globalTag = "GLOBAL::BALLS",
        writeTiers = ['RECO'])

    rerecoCmsswHelper.addOutputModule(
        "outputRECO", primaryDataset = "Primary",
        processedDataset = "Processed",
        dataTier = "RECO")

    #Add a stageOut step
    #skimStageOutHelper = skimStageOut.getTypeHelper()
    #skimLogArchHelper  = skimLogArch.getTypeHelper()


    rereco.addGenerator("BasicNaming")
    workload.setSiteWhitelist(["T2_XX_SiteA", "T2_XX_SiteB", "T2_XX_SiteC"])
    return workload



"""
_ExistWorker_

Oracle implementation of ExistWorker
"""

__all__ = []



from WMCore.Agent.Database.MySQL.ExistWorker import ExistWorker \
     as ExistWorkerMySQL

class ExistWorker(ExistWorkerMySQL):
    pass



"""
_CreateAgent_

Implementation of CreateAgent for MySQL.

Inherit from CreateAgentBase, and add MySQL specific substitutions (e.g. add
INNODB).
"""

from WMCore.Agent.Database.CreateAgentBase import CreateAgentBase


class Create(CreateAgentBase):
    """
    Class to set up the Agent schema in a MySQL database
    """

    def __init__(self, logger=None, dbi=None, params=None):
        """
        _init_

        Call the base class's constructor and create all necessary tables,
        constraints and inserts.
        """
        CreateAgentBase.__init__(self, logger, dbi, params)

    def execute(self, conn=None, transaction=None):
        for i in self.create:
            self.create[i] += " ENGINE=InnoDB ROW_FORMAT=DYNAMIC"
            self.create[i] = self.create[i].replace('INTEGER', 'INT(11)')

        return CreateAgentBase.execute(self, conn, transaction)



#!/usr/bin/env python

"""
Oracle implementation of AddChecksum
"""





from WMComponent.DBS3Buffer.MySQL.DBSBufferFiles.AddChecksum import AddChecksum as MySQLAddChecksum

class AddChecksum(MySQLAddChecksum):
    """
    Identical to MySQL Version

    """



#!/usr/bin/env python
"""
_Exists_

Oracle implementation of Jobs.Exists
"""

__all__ = []



from WMCore.WMBS.MySQL.Jobs.Exists import Exists as ExistsJobMySQL

class Exists(ExistsJobMySQL):
    sql = ExistsJobMySQL.sql



#!/usr/bin/env python
"""
_CMSSW_

Template for a CMSSW Step

"""

import pickle

from future.utils import viewitems

from Utils.Utilities import encodeUnicodeToBytes
from WMCore.WMSpec.ConfigSectionTree import nodeName
from WMCore.WMSpec.Steps.Template import CoreHelper, Template
from WMCore.Cache.WMConfigCache import ConfigCacheException
import shlex


class CMSSWStepHelper(CoreHelper):
    """
    _CMSSWStepHelper_

    Add API calls and helper methods to the basic WMStepHelper to specialise
    for CMSSW tasks

    """
    def setAcqEra(self, acqEra):
        """
        _setAcqEra_
        Set the acquisition era attribute for this step.
        """
        self.data.output.acqEra = acqEra

    def setProcStr(self, procStr):
        """
        _setProcStr_
        Set the processing string attribute for this step.
        """
        self.data.output.procStr = procStr

    def setProcVer(self, procVer):
        """
        _setProcVer_
        Set the processing version era attribute for this step.
        """
        self.data.output.procVer = procVer

    def getAcqEra(self):
        """
        _getAcqEra_
        Retrieve the acquisition era for this step, or return None if non-existent.
        """
        return getattr(self.data.output, 'acqEra', None)

    def getProcStr(self):
        """
        _getProcStr_
        Retrieve the processing string for this step, or return None if non-existent.
        """
        return getattr(self.data.output, 'procStr', None)

    def getProcVer(self):
        """
        _getProcVer_
        Retrieve the processing version for this step, or return None if non-existent.
        """
        return getattr(self.data.output, 'procVer', None)

    def setPrepId(self, prepId):
        """
        _setPrepId_
        Set the prep_id attribute for this step.
        """
        self.data.output.prepId = prepId

    def getPrepId(self):
        """
        _getPrepId_
        Retrieve the prep_id for this step, or return None if non-existent.
        """
        return getattr(self.data.output, 'prepId', None)

    def addOutputModule(self, moduleName, **details):
        """
        _addOutputModule_

        Add in an output module settings, all default to None unless
        the value is provided in details

        """
        modules = self.data.output.modules

        if getattr(modules, moduleName, None) == None:
            modules.section_(moduleName)
        module = getattr(modules, moduleName)

        for key, value in viewitems(details):
            setattr(module, key, value)

        return

    def listOutputModules(self):
        """
        _listOutputModules_

        retrieve list of output module names

        """
        if hasattr(self.data.output, "modules"):
            return list(self.data.output.modules.dictionary_())

        return []

    def getOutputModule(self, name):
        """
        _getOutputModule_

        retrieve the data structure for an output module by name
        None if not found

        """
        return getattr(self.data.output.modules, name, None)

    def setConfigCache(self, url, document, dbName="config_cache"):
        """
        _setConfigCache_

        Set the information required to retrieve a configuration from
        the config cache.

        url - base URL for the config cache instance
        document - GUID for the config document
        dbName - optional, name of the db instance in the couch server

        """
        self.data.application.configuration.configCacheUrl = url
        self.data.application.configuration.configId = document
        self.data.application.configuration.cacheName = dbName
        docUrl = "%s/%s/%s" % (url, dbName, document)
        self.data.application.configuration.configUrl = docUrl
        self.data.application.configuration.retrieveConfigUrl = "%s/configFile" % docUrl

    def setDataProcessingConfig(self, scenarioName, functionName, **args):
        """
        _setDataProcessingConfig_

        Set a configuration library to be used from the CMSSW Release
        DataProcessing package.

        """
        self.data.application.configuration.scenario = scenarioName
        self.data.application.configuration.function = functionName
        # assume if this crashes we are dealing with complex data
        # which is only supported in new agents that only look
        # at pickledarguments anyways
        try:
            self.data.application.configuration.section_('arguments')
            [setattr(self.data.application.configuration.arguments, k, v) for k, v in viewitems(args)]
        except Exception:
            pass
        # FIXME: once both central services and WMAgent are in Py3, we can remove protocol=0
        self.data.application.configuration.pickledarguments = pickle.dumps(args, protocol=0)
        return

    def cmsswSetup(self, cmsswVersion, **options):
        """
        _cmsswSetup_

        Provide setup details for CMSSW.

        cmsswVersion - required - version of CMSSW to use

        Optional:

        scramCommand - defaults to scramv1
        scramProject - defaults to CMSSW
        scramArch    - optional scram architecture, defaults to None
        buildArch    - optional scram build architecture, defaults to None
        softwareEnvironment - setup command to bootstrap scram,defaults to None
        """
        self.data.application.setup.cmsswVersion = cmsswVersion
        for k, v in viewitems(options):
            setattr(self.data.application.setup, k, v)
        return

    def getScramArch(self):
        """
        _getScramArch_

        Retrieve the scram architecture used for this step.
        """
        return self.data.application.setup.scramArch

    def getCMSSWVersion(self):
        """
        _getCMSSWVersion_

        Retrieve the version of the framework used for this step.
        """
        return self.data.application.setup.cmsswVersion

    def setGlobalTag(self, globalTag):
        """
        _setGlobalTag_

        Set the global tag.
        """
        self.data.application.configuration.section_('arguments')
        self.data.application.configuration.arguments.globalTag = globalTag

        args = {}
        if hasattr(self.data.application.configuration, "pickledarguments"):
            args = pickle.loads(encodeUnicodeToBytes(self.data.application.configuration.pickledarguments))
        args['globalTag'] = globalTag
        # FIXME: once both central services and WMAgent are in Py3, we can remove protocol=0
        self.data.application.configuration.pickledarguments = pickle.dumps(args, protocol=0)

        return

    def getGlobalTag(self):
        """
        _getGlobalTag_

        Retrieve the global tag.
        """
        if hasattr(self.data.application.configuration, "arguments"):
            if hasattr(self.data.application.configuration.arguments, "globalTag"):
                return self.data.application.configuration.arguments.globalTag

        pickledArgs = encodeUnicodeToBytes(self.data.application.configuration.pickledarguments)
        return pickle.loads(pickledArgs)['globalTag']

    def setDatasetName(self, datasetName):
        """
        _setDatasetName_

        Set the dataset name in the pickled arguments
        """
        self.data.application.configuration.section_('arguments')
        self.data.application.configuration.arguments.datasetName = datasetName

        args = {}
        if hasattr(self.data.application.configuration, "pickledarguments"):
            args = pickle.loads(encodeUnicodeToBytes(self.data.application.configuration.pickledarguments))
        args['datasetName'] = datasetName
        # FIXME: once both central services and WMAgent are in Py3, we can remove protocol=0
        self.data.application.configuration.pickledarguments = pickle.dumps(args, protocol=0)

        return

    def getDatasetName(self):
        """
        _getDatasetName_

        Retrieve the dataset name from the pickled arguments
        """
        if hasattr(self.data.application.configuration, "arguments"):
            if hasattr(self.data.application.configuration.arguments, "datasetName"):
                return self.data.application.configuration.arguments.datasetName

        return pickle.loads(self.data.application.configuration.pickledarguments).get('datasetName', None)

    def getScenario(self):
        """
        _getScenario_

        Retrieve the scenario from the pickled arguments, if any
        """
        if hasattr(self.data.application.configuration, "scenario"):
            return self.data.application.configuration.scenario

        return None

    def setUserSandbox(self, userSandbox):
        """
        _setUserSandbox_

        Sets the userSandbox. Eventually may have to move this to a proper
        list rather than a one element list
        """
        if userSandbox:
            self.data.user.inputSandboxes = [userSandbox]
        return

    def setUserFiles(self, userFiles):
        """
        _setUserFiles_

        Sets the list of extra files the user needs
        """
        if userFiles:
            self.data.user.userFiles = userFiles
        return

    def setUserLFNBase(self, lfnBase):
        """
        _setUserFiles_

        Sets the list of extra files the user needs
        """
        if lfnBase:
            self.data.user.lfnBase = lfnBase
        return

    def setupChainedProcessing(self, inputStepName, inputOutputModule):
        """
        _setupChainedProcessing_

        Set values to support chained CMSSW running.
        """
        self.data.input.chainedProcessing = True
        self.data.input.inputStepName = inputStepName
        self.data.input.inputOutputModule = inputOutputModule

    def keepOutput(self, keepOutput):
        """
        _keepOutput_

        Mark whether or not we should keep the output from this step.  We don't
        want to keep the output from certain chained steps.
        """
        self.data.output.keep = keepOutput
        return

    def getPileup(self):
        """
        _getPileup_

        Retrieve the pileup config from this step.
        """
        return getattr(self.data, "pileup", None)

    def setupPileup(self, pileupConfig, dbsUrl):
        """
        include pileup input configuration into this step configuration.
        pileupConfig is initially specified as input to the workload
        (user input) and here is available as a dict.

        """
        # so, e.g. this {"cosmics": "/some/cosmics/dataset", "minbias": "/some/minbias/dataset"}
        # would translate into
        # self.data.pileup.comics.dataset = "/some/cosmics/dataset"
        # self.data.pileup.minbias.dataset = "/some/minbias/dataset"
        self.data.section_("pileup")
        for pileupType, dataset in viewitems(pileupConfig):
            self.data.pileup.section_(pileupType)
            setattr(getattr(self.data.pileup, pileupType), "dataset", dataset)
        setattr(self.data, "dbsUrl", dbsUrl)

    def setEventsPerLumi(self, eventsPerLumi):
        """
        _setEventsPerLumi_
        Add event per lumi information to the step, so it can be added later
        to the process, this comes from user input
        """
        if eventsPerLumi != None:
            setattr(self.data.application.configuration, "eventsPerLumi", eventsPerLumi)

    def getSkipBadFiles(self):
        """
        _getSkipBadFiles_

        Check if we can skip inexistent files instead of failing the job
        """
        return getattr(self.data.application.configuration, "skipBadFiles", False)

    def setSkipBadFiles(self, skipBadFiles):
        """
        _setSkipBadFiles_

        Add a flag to indicate the CMSSW process if we can
        skip inexistent files instead of failing the job
        """
        setattr(self.data.application.configuration, "skipBadFiles", skipBadFiles)

    def setNumberOfCores(self, ncores, nEventStreams=0):
        """
        _setNumberOfCores_

        Set the number of cores and event streams for CMSSW to run on
        """
        # if None is passed for EventStreams, then set it to 0
        nEventStreams = nEventStreams or 0

        self.data.application.multicore.numberOfCores = ncores
        self.data.application.multicore.eventStreams = nEventStreams

    def getNumberOfCores(self):
        """
        _getNumberOfCores_

        Get number of cores
        """
        return self.data.application.multicore.numberOfCores

    def getEventStreams(self):
        """
        _getEventStreams_

        Get number of event streams
        """
        return self.data.application.multicore.eventStreams

    def setGPUSettings(self, requiresGPU, gpuParams):
        """
        Set whether this CMSSW step should require GPUs and if so, which
        setup should be allowed and/or used
        """
        self.data.application.gpu.gpuRequired = requiresGPU
        self.data.application.gpu.gpuRequirements = gpuParams

    def getGPUSettings(self):
        """
        Return the GPU configuration for this CMSSW step
        :return: a tuple with:
          * string whether GPU is required or not
          * dictionary with the GPU requirements (or None)
        """
        return (self.data.application.gpu.gpuRequired,
                self.data.application.gpu.gpuRequirements)

    def getPhysicsTypeFromStepsArg(self, stepsArg, hasPileup=False, hasDatamix=False):
        """
        Get information about the step physics type, based on the
        cmsDriver command line --steps argument
        Reference: https://github.com/cms-sw/cmssw/issues/42587
        :param stepsArg: str with the cmsDriver step arguments
        :param hasPileup: bool, cmsDriver option
        :param hasDatamix: bool, cmsDriver option
        :return: str with all physics types, comman separated
        """
        physicsTypes = {"GEN":"GEN", "SIM":"SIM", "DIGI":"DIGI", "RECO":"RECO", "PAT":"MINIAOD", "NANO":"NANOAOD", "ALL":"GEN,SIM,DIGI,RECO"}
        stepTypes = []
        for stepType in physicsTypes.keys():
            for step in stepsArg.split(","):
                # Steps like DIGI could come in the form of DIGI:pdigi_valid
                step = step.split(':')[0]
                if step == stepType:
                    physicsType = physicsTypes.get(stepType)
                    if step == "DIGI":
                        if hasPileup:
                            if hasDatamix: 
                                physicsType = "DIGI_premix"
                            else:
                                physicsType = "DIGI_classicalmix"
                        else:
                            physicsType = "DIGI_nopileup"
                    stepTypes.append(physicsType)
        if not stepTypes:
            stepTypes.append("UNKNOWN")

        return ",".join(stepTypes)

    def getCmsDriverCommandLineArgs(self, conf):
        """
        Find the command line arguments in a pset config file
        :param conf: config document from configCache object
        :return: str with the cmsDriver command line arguments
        """
        # First, split all new lines
        conf = conf.split('\n')
        # Comment line 4 usually has the command line arguments
        # but find it otherwise (from the comment lines in the conf)
        cmdPattern = "# with command line options:"
        cmd = None
        if conf[4].startswith(cmdPattern):
            cmd = conf[4]
        else:
            for line in conf:
                if line.startswith(cmdPattern):
                    cmd = line
                elif not line.startswith('#'):
                    break
        # Replace "=" with spaces
        # E.g.: --step=DIGI -> --step DIGI
        if cmd:
            cmd = cmd.replace("=", " ")
    
        return cmd

    def determinePhysicsType(self, configCache=None):
        """
        Get information about the step physics type
        :param configCache: configCache object
        :return: str with the physics types, command separated.
        """
        hasPileup = False
        hasDatamix = False
        try:
            if configCache:
                # command arguments are in line 4
                myconf = configCache.getConfig()
                if myconf == None:
                    return "UNKNOWN"
                cmd = self.getCmsDriverCommandLineArgs(myconf)
                cmd = shlex.split(cmd)
                # If we are processing data,
                # just return "DataProcessing"
                if "--data" in cmd:
                    return "DataProcessing"
                # The command step arg is either '--step' or '-s'
                # Check for premixing and pileup parameters
                if "--pileup_input" in cmd:
                    hasPileup = True
                if "--datamix" in cmd:
                    hasDatamix = True
                try:
                    stepsArg = cmd[cmd.index("--step")+1]
                except ValueError:
                    stepsArg = cmd[cmd.index("-s")+1]
            else:
                stepsArg = "UNKNOWN"
        except (ConfigCacheException, AttributeError, ValueError) as ex:
            stepsArg = "UNKNOWN"

        return self.getPhysicsTypeFromStepsArg(stepsArg, hasPileup, hasDatamix)

    def getPhysicsType(self):
        """
        Set the physics type of the step
        :return: str
        """
        return getattr(self.data, "stepPhysicsType", None)

    def setPhysicsType(self, configCache=None):
        """
        Set the physics type of the step.
        :param configCache: configCache object
        """
        self.data.stepPhysicsType = self.determinePhysicsType(configCache)


class CMSSW(Template):
    """
    _CMSSW_

    Tools for creating a template CMSSW Step

    """

    def install(self, step):
        """
        _install_

        Add the set of default fields to the step required for running
        a cmssw job

        """
        stepname = nodeName(step)
        step.stepType = "CMSSW"
        step.application.section_("setup")
        step.application.setup.scramCommand = "scramv1"
        step.application.setup.scramProject = "CMSSW"
        step.application.setup.cmsswVersion = None
        step.application.setup.scramArch = None
        step.application.setup.buildArch = None
        step.application.setup.softwareEnvironment = None

        step.application.section_("command")
        step.application.command.executable = "cmsRun"
        step.application.command.configuration = "PSet.py"
        step.application.command.configurationPickle = "PSet.pkl"
        step.application.command.configurationHash = None
        step.application.command.psetTweak = None
        step.application.command.arguments = ""
        step.output.jobReport = "FrameworkJobReport.xml"
        step.output.stdout = "%s-stdout.log" % stepname
        step.output.stderr = "%s-stderr.log" % stepname
        step.output.keep = True
        step.output.section_("modules")

        step.output.section_("analysisFiles")

        step.section_("runtime")
        step.runtime.preScripts = []
        step.runtime.scramPreScripts = []
        step.runtime.postScripts = []
        step.runtime.postScramScripts = []

        step.section_("debug")
        step.debug.verbosity = 0
        step.debug.keepLogs = False

        step.section_("user")
        step.user.inputSandboxes = []
        step.user.script = None
        step.user.outputFiles = []
        step.user.userFiles = []
        step.user.lfnBase = None

        step.section_("monitoring")

        # support for multicore cmssw running mode
        step.application.section_("multicore")
        step.application.multicore.numberOfCores = 1
        step.application.multicore.eventStreams = 0

        # support for GPU in CMSSW (using defaults from StdBase)
        step.application.section_("gpu")
        step.application.gpu.gpuRequired = "forbidden"
        step.application.gpu.gpuRequirements = None

    def helper(self, step):
        """
        _helper_

        Wrap the WMStep provided in the CMSSW helper class that
        includes the ability to add and manipulate the details
        of a CMSSW workflow step

        """
        return CMSSWStepHelper(step)



#!/usr/bin/env python
"""
WorkQueue Policy

https://twiki.cern.ch/twiki/bin/view/CMS/WMCoreWorkQueuePolicy
"""


#
#from WMCore.WMFactory import WMFactory
#from WMCore.WMSpec.WMWorkload import WMWorkloadHelper
#
#startFac = WMFactory("WorkQueueStartPolicy", __name__ + '.Start')
#endFac = WMFactory("WorkQueueEndPolicy", __name__ + '.End')
#startMap = {}
#
##TODO: Argument handling need to give params from workqueue as well as from spec
#
#def setStartPolicy(policyDirective, policyName, policyArgs):
#    """Set the Start policy for a queue"""
#    global startMap
#    startMap[policyDirective] = (policyName, policyArgs)
#
#def startPolicy(name, startMap):
#    """Load a start policy"""
#    # Do we have a defined policy for this policy - if not load null policy
#    if not startMap.has_key(name):
#        name = 'passthrough'
#    policyName, args = startMap[name]
#    #policyName, args = startMap.get(name, ('passthrough', {}))
#    return startFac.loadObject(policyName,
#                               args,
#                               storeInCache = False,
#                               )
##    return startFac.loadObject(startMap.get(name, name),
##                               args,
##                               storeInCache = False)
#
#def endPolicy(elements, args):
#    """Load an end policy"""
#
#    for ele in elements:
#        # if still working return simple aggregate state
#        if not ele.inEndState():
#            return endFac.loadObject('SingleShot',
#                             {'SuccessThreshold' : 1.0},
#                             storeInCache = False)(*elements)
#
#    # all elements finished processing load policy and apply
#    spec = WMWorkloadHelper()
#    spec.load(elements[0]['WMSpecUrl'])
#    args.update(spec.data.policies.end.dictionary_())
#    return endFac.loadObject(spec.data.policies.end.policyName,
#                             args,
#                             storeInCache = False)(*elements)
#
#__all__ = [setStartPolicy, startPolicy, endPolicy]



#!/usr/bin/env python
"""
_LoadForTaskArchiver_

Oracle implementation of Jobs.LoadForTaskArchiver

Created on Sep 5, 2012

@author: dballest
"""

from WMCore.WMBS.MySQL.Jobs.LoadForTaskArchiver import LoadForTaskArchiver as MySQLLoadForTaskArchiver

class LoadForTaskArchiver(MySQLLoadForTaskArchiver):
    pass



#!/usr/bin/env python
"""
_AddLocation_

Oracle implementation of DBSBufferFiles.AddLocation
"""

from WMComponent.DBS3Buffer.MySQL.DBSBufferFiles.AddLocation import AddLocation as MySQLAddLocation

class AddLocation(MySQLAddLocation):

    sql = """INSERT INTO dbsbuffer_location
             (id, pnn)
             SELECT dbsbuffer_location_seq.nextval, :location
             FROM DUAL
             WHERE NOT EXISTS
               ( SELECT *
                 FROM dbsbuffer_location
                 WHERE pnn = :location )
             """



#!/usr/bin/env python
"""
_GetJobGroups_

Oracle implementation of Subscriptions.GetJobGroups
"""




from WMCore.WMBS.MySQL.Subscriptions.GetJobGroups import GetJobGroups as GetJobGroupsMySQL

class GetJobGroups(GetJobGroupsMySQL):
    pass



#!/usr/bin/env python
# encoding: utf-8
"""
ReproducibleSeeding.py

Created by Dave Evans on 2010-08-30.
Copyright (c) 2010 Fermilab. All rights reserved.

Seeding algorithm that generates job seeds at splitting time and saves them as part of the job itself, so that
the seeds are the same every time the job gets run (used for RelVal)

List of seeds is retrieved from the ConfigCache PSet Tweak

"""


from WMCore.JobSplitting.Generators.GeneratorInterface import GeneratorInterface
from WMCore.Cache.WMConfigCache import ConfigCache
from WMCore.Configuration import ConfigSection

import random
from random import SystemRandom
_inst  = SystemRandom()

#  //
# // Max bit mask size for 32 bit integers
#//  This is required to be max 900M for CMSSW
_MAXINT = 900000000
newSeed = lambda : _inst.randint(1, _MAXINT)


class ReproducibleSeeding(GeneratorInterface):
    def __init__(self, **options):
        GeneratorInterface.__init__(self, **options)
        self.couchUrl = options.get("CouchUrl")
        self.couchDBName = options.get("CouchDBName")
        self.couchConfigDoc = options.get("ConfigCacheDoc")

        confCache = ConfigCache(dbURL = self.couchUrl, couchDBName = self.couchDBName, id = self.couchConfigDoc)
        confCache.load()
        seeds = confCache.document[u'pset_tweak_details'][u'process'][u'RandomNumberGeneratorService']
        self.seedTable = []
        for k in seeds:
            if k == u"parameters_" : continue
            self.seedTable.append("process.RandomNumberGeneratorService.%s.initialSeed" % k)


    def __call__(self, wmbsJob):
        wmbsJob.addBaggageParameter("seeding", self.__class__.__name__)

        for seed in self.seedTable:
            wmbsJob.addBaggageParameter(seed, newSeed())



#!/usr/bin/env python
"""
_HeritageLFNChild_

MySQL implementation of DBSBufferFiles.HeritageLFNChild
"""




from WMCore.Database.DBFormatter import DBFormatter

class HeritageLFNChild(DBFormatter):
    sql = """INSERT INTO dbsbuffer_file_parent (child, parent)
               SELECT dbsbuffer_file.id, :parent FROM dbsbuffer_file
                 WHERE dbsbuffer_file.lfn = :lfn"""

    def execute(self, childLFNs, parentID, conn = None, transaction = False):
        binds = []
        for childLFN in childLFNs:
            binds.append({"lfn": childLFN, "parent": parentID})

        self.dbi.processData(self.sql, binds, conn = conn,
                             transaction = transaction)
        return



#!/usr/bin/env python
"""
_DaemonDetails_

Util object to parse the Daemon parameters for a component and
convert it into a dictionary

Also, provides utils to shutdown the daemon process

"""
from __future__ import print_function

from builtins import str, range

import os
import subprocess
import shutil
import time
# FIXME: needs to be replaced with persistent backend.
from xml.dom.minidom import parse

from Utils.PythonVersion import PY3
from Utils.Utilities import encodeUnicodeToBytesConditional


def run(command):
    proc = subprocess.Popen(
        ["/bin/bash"], shell=True, cwd=os.environ['PWD'],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdin=subprocess.PIPE,
        )

    proc.stdin.write(encodeUnicodeToBytesConditional(command, condition=PY3))
    stdout, stderr = proc.communicate()
    rc = proc.returncode

    return stdout, stderr, rc


class Details(dict):
    """
    _DaemonDetails_

    Dictionary of information about a Component Daemon
    along with some utility methods to manage it

    Components, upon startup, produce a Daemon.xml file in their component
    directory containing the details of the daemon process.
    This util takes that file, reads its fields into a dictionary.

    """

    def __init__(self, daemonXmlFile):
        dict.__init__(self)
        self.load(daemonXmlFile)
        self.daemonXmlFile = daemonXmlFile

    def load(self, xmlFile):
        """
        _load_

        Load the XML values into this instance

        """
        dom = parse(xmlFile)
        daemons = dom.getElementsByTagName("Daemon")
        for daemon in daemons:
            childNodes = daemon.childNodes
            for childNode in childNodes:
                name = childNode.nodeName
                if name != "#text":
                    value = childNode.getAttribute("Value")
                    self[name] = int(value)

    def isAlive(self):
        """
        _isAlive_

        Is the process still running?
        """
        # Reference: ps -T -p 1946167 -o euser,pid,ppid,lwp,nlwp,stat,start
        # it prints the user, process and its threads, number of threads, etc
        dummyse, dummyso, rc = run('ps -p %s' % self['ProcessID'])
        if rc != 0:
            return False
        return True

    def kill(self, signal=15):
        """
        _kill_

        Kill the process with the signal provided

        """
        os.kill(self['ProcessID'], signal)
        time.sleep(1)
        self.removeAndBackupDaemonFile()
        return

    def killWithPrejudice(self, signal=15):
        """
        _killWithPredjudice_

        Issue the kill, then watch to make sure it shuts down.
        If it takes more than a couple of seconds, kill -9 it.

        """
        os.killpg(self['ProcessGroupID'], signal)
        for dummycount in range(0, 3):
            time.sleep(1)
            if not self.isAlive():
                self.removeAndBackupDaemonFile()
                return
            continue
        os.killpg(self['ProcessGroupID'], 9)
        self.removeAndBackupDaemonFile()
        return

    def removeAndBackupDaemonFile(self):
        """
        Removes the daemon file (after a kill) and backs it up
        for post mortem.
        """
        path, dFile = os.path.split(self.daemonXmlFile)
        timeStamp = time.strftime("%d-%M-%Y")
        newFile = "%s.BAK.%s" % (dFile, str(timeStamp))
        newLocation = os.path.join(path, newFile)
        try:
            shutil.move(self.daemonXmlFile, newLocation)
        except Exception as ex:
            print('Move failed. Remove manual: ' + str(ex))



# -*- coding: utf-8 -*-
"""
_SetRunningSlots_

Oracle implementation of Locations.SetRunningSlots
Created on Mon Jun 18 14:08:58 2012

@author: dballest
"""

from WMCore.WMBS.MySQL.Locations.SetRunningSlots import SetRunningSlots as MySQLSetRunningSlots

class SetRunningSlots(MySQLSetRunningSlots):
    """
    Same as MySQL version
    """
    pass



#!/usr/bin/env python
"""
MySQL implementation of File.Heritage

Make the parentage link between two file id's
"""
__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class Heritage(DBFormatter):
    sql = """insert into wmbs_file_parent (child, parent) values (:child, :parent)"""

    def getBinds(self, parent=0, child=0):
        # Can't use self.dbi.buildbinds here...
        binds = []

        parent = self.dbi.makelist(parent)
        child = self.dbi.makelist(child)
        for p in parent:
            for c in child:
                binds.append({'child': c,
                              'parent': p})
        return binds

    def format(self, result):
        return True

    def execute(self, parent=0, child=0, conn = None, transaction = False):
        binds = self.getBinds(parent, child)

        self.logger.debug('File.Heritage binds: %s' % binds)
        result = self.dbi.processData(self.sql, binds,
                         conn = conn, transaction = transaction)
        return self.format(result)



#!/usr/bin/env python
"""
_SetBlocksRule_

Oracle implementation of SetBlocksRule
"""

from __future__ import division

from WMComponent.RucioInjector.Database.MySQL.SetBlocksRule import SetBlocksRule as MySQLBase


class SetBlocksRule(MySQLBase):
    pass



#!/usr/bin/env python
"""
MySQL implementation of Jobs.LoadOutputID

Load the ID of the output fileset for a job
"""
__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class LoadOutputID(DBFormatter):
    """

    Load the ID of the output fileset for a job

    """


    sql = """SELECT wfs.id AS id FROM wmbs_fileset wfs
                INNER JOIN wmbs_jobgroup wjg ON wjg.output = wfs.id
                INNER JOIN wmbs_job wj ON wj.jobgroup = wjg.id
                WHERE wj.id = :jobid"""


    def execute(self, jobID, conn = None, transaction = False):
        """
        Given a jobID, find the fileset

        """

        result = self.dbi.processData(self.sql, {"jobid": jobID}, conn = conn,
                                      transaction = transaction)
        return self.formatDict(result)[0].get('id', None)



#!/usr/bin/env python
"""WorkQueue Exceptions"""

class WorkQueueError(Exception):
    """Standard error baseclass"""
    def __init__(self, error):
        Exception.__init__(self, error)
        self.msg = WorkQueueError.__class__.__name__
        self.error = error

    def __str__(self):
        return "%s: %s" % (self.msg, self.error)

class WorkQueueWMSpecError(WorkQueueError):
    """Problem with the spec file"""
    def __init__(self, wmspec, error):
        WorkQueueError.__init__(self, error)
        self.wmspec = wmspec
        if hasattr(self.wmspec, 'name'):
            self.msg = "Invalid WMSpec: '%s'" % self.wmspec.name()
        else:
            self.msg = "Invalid WMSpec:"

class WorkQueueNoWorkError(WorkQueueError):
    """No work for spec"""
    def __init__(self, wmspec, error):
        WorkQueueError.__init__(self, error)
        self.wmspec = wmspec
        if hasattr(self.wmspec, 'name'):
            self.msg = "No work in spec: '%s' Check inputs" % self.wmspec.name()
        else:
            self.msg = "No work in spec: Check inputs"

class WorkQueueNoMatchingElements(WorkQueueError):
    """Didn't find any element"""
    def __init__(self, error):
        WorkQueueError.__init__(self, error)
        self.msg = WorkQueueNoMatchingElements.__class__.__name__
        self.error = error

    def __str__(self):
        return "%s: %s" % (self.msg, self.error)

TERMINAL_EXCEPTIONS = (WorkQueueWMSpecError, WorkQueueNoWorkError)



#!/usr/bin/env python
# encoding: utf-8
"""
TestInitCouchApp.py

Specialisation of TestInit for Tests that require a couch connection.

Own thing to avoid making entire Test stack depend on code from couchapp

Created by Dave Evans on 2010-08-19.
Copyright (c) 2010 Fermilab. All rights reserved.
"""

from builtins import object
from future import standard_library

standard_library.install_aliases()

import os
import urllib.parse
import logging

from couchapp.commands import push as couchapppush
from WMCore.Database.CMSCouch import CouchServer

from WMQuality.TestInit import TestInit


class CouchAppTestHarness(object):
    """
    Test Harness for installing a couch database instance with several couchapps
    in a unittest.setUp and wiping it out in a unittest.tearDown


    """

    def __init__(self, couchUrl=None, testClassName=None):
        self.couchUrl = os.environ.get("COUCHURL", couchUrl)
        if self.couchUrl == None:
            msg = "COUCHRURL env var not set..."
            raise RuntimeError(msg)
        if self.couchUrl.endswith('/'):
            raise RuntimeError("COUCHURL env var shouldn't end with /")
        self.couchServer = CouchServer(self.couchUrl)
        self.testClassName = testClassName
        self.logger = logging.getLogger()

    def create(self, dbName, dropExistingDb=True):
        """
        create couch db instance

        :param dbName: database name to create
        :param dropExistingDb: flag to drop existing database name if it exists
        """
        # import pdb
        # pdb.set_trace()
        databases = self.couchServer.listDatabases()
        msg = f"Create {dbName} in {self.couchUrl}, existing DBs={databases}"
        self.logger.info(msg)
        if dbName in databases:
            if not dropExistingDb:
                return
            self.drop(dbName)

        self.couchServer.createDatabase(dbName)

    def drop(self, dbName):
        """
        drop given database name from CouchDB

        :param dbName: database name
        """
        databases = self.couchServer.listDatabases()
        msg = f"Drop {dbName} in {self.couchUrl}, existing DBs={databases}"
        self.logger.info(msg)
        if dbName in databases:
            self.couchServer.deleteDatabase(dbName)

    def pushCouchapps(self, dbName, *couchappdirs):
        """
        push a list of couchapps to the given database

        :param dbName: database name
        :param couchappdirs: list of couch applications
        """
        for couchappdir in couchappdirs:
            msg = f"Push {couchappdir} in {self.couchUrl} database {dbName}"
            self.logger.info(msg)
            couchapppush(couchappdir, "%s/%s" % (self.couchUrl, urllib.parse.quote_plus(dbName)))


class TestInitCouchApp(TestInit):
    """
    TestInit with baked in Couch goodness
    """

    def __init__(self, testClassName, dropExistingDb=True):
        TestInit.__init__(self, testClassName)
        self.databases = []
        self.couch = CouchAppTestHarness(testClassName=testClassName)
        # for experiments, after tests run, it's useful to have CouchDB
        # populated with the testing data - having tearDownCouch commented
        # out, this flag prevents from re-initializing the database
        self.dropExistingDb = dropExistingDb
        self.couchUrl = self.couch.couchUrl
        self.couchDbName = "Not set yet"

    def couchAppRoot(self, couchapp):
        """
        Return parent path containing couchapp

        :param couchapp: couch application to use
        """
        wmcoreroot = os.path.normpath(os.path.join(self.init.getWMBASE(), '..', '..', '..'))
        develPath = os.path.join(self.init.getWMBASE(), "src", "couchapps")
        if os.path.exists(os.path.join(develPath, couchapp)):
            return develPath
        elif os.path.exists(os.path.join(wmcoreroot, 'xdata', 'couchapps', couchapp)):
            return os.path.join(wmcoreroot, 'xdata', 'couchapps')
        elif os.path.exists(os.path.join(wmcoreroot, 'data', 'couchapps', couchapp)):
            return os.path.join(wmcoreroot, 'data', 'couchapps')
        raise OSError('Cannot find couchapp: %s' % couchapp)

    def setupCouch(self, dbName, *couchapps):
        """
        _setupCouch_

        Call in the setUp of your test to build a couch instance with the dbname provided
        and the required list of couchapps from WMCore/src/couchapps

        :param dbName: database name
        :param *couchapps: list of couch apps associated with given database
        """
        # this function performs two set of actions:
        # 1. it creates given dbName
        # 2. it registter given set of couchapps in that DB
        # since those are independent operations we separate them here to avoid
        # racing conditions on multiple setupCouch() calls with the same dbName but different apps, e.g.
        # see test/python/WMCore_t/WMSpec_t/StdSpecs_t/Resubmission_t.py
        #     self.testInit.setupCouch("resubmission_t", "ReqMgr")
        #     self.testInit.setupCouch("resubmission_t", "ConfigCache")
        if dbName not in self.databases:
            self.databases.append(dbName)
            self.couch.create(dbName, dropExistingDb=self.dropExistingDb)
        # just create the db is couchapps are not specified
        if len(couchapps) > 0:
            self.couch.pushCouchapps(dbName, *[os.path.join(self.couchAppRoot(couchapp), couchapp) for couchapp in couchapps])
        self.couchDbName = dbName

    def tearDownCouch(self):
        """
        _tearDownCouch_

        call this in tearDown to erase all evidence of your couch misdemeanours
        """
        for database in self.databases:
            self.couch.drop(database)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Version of WMCore/Services/Rucio intended to be used with mock or unittest.mock
"""
from __future__ import print_function, division
# from builtins import object # avoid importing this, it beraks things

import json
import logging
import os
import hashlib

from WMCore.Services.DBS.DBS3Reader import DBS3Reader, DBSReaderError
from WMCore.Services.Rucio.Rucio import WMRucioException, WMRucioDIDNotFoundException
from WMCore.WMBase import getTestBase
from WMQuality.Emulators.DataBlockGenerator.DataBlockGenerator import DataBlockGenerator

from Utils.PythonVersion import PY2, PY3
from Utils.Utilities import encodeUnicodeToBytesConditional

PROD_DBS = 'https://cmsweb-prod.cern.ch/dbs/prod/global/DBSReader'

NOT_EXIST_DATASET = 'thisdoesntexist'
# PILEUP_DATASET = '/HighPileUp/Run2011A-v1/RAW'
PILEUP_DATASET = '/GammaGammaToEE_Elastic_Pt15_8TeV-lpair/Summer12-START53_V7C-v1/GEN-SIM'

SITES = ['T2_XX_SiteA', 'T2_XX_SiteB', 'T2_XX_SiteC']
_BLOCK_LOCATIONS = {}

BLOCKS_PER_DATASET = 2
FILES_PER_BLOCK = 5
FILES_PER_DATASET = BLOCKS_PER_DATASET * FILES_PER_BLOCK

mockFile = os.path.join(getTestBase(), '..', 'data', 'Mock', 'RucioMockData.json')
try:
    with open(mockFile, 'r') as jsonObj:
        MOCK_DATA = json.load(jsonObj)
except IOError:
    MOCK_DATA = {}


class MockRucioApi(object):
    """
    Version of WMCore/Services/Rucio intended to be used with mock or unittest.mock
    """

    def __init__(self, acct, hostUrl=None, authUrl=None, configDict=None):
        print("Using MockRucioApi: acct={}, url={}, authUrl={}".format(acct, hostUrl, authUrl))
        configDict = configDict or {}
        self.dataBlocks = DataBlockGenerator()
        self.subRequests = {}
        self.rucioParams = {}

    def sitesByBlock(self, block):
        """
        Centralize the algorithm to decide where a block is based on the hash name
        :param block: the name of the block
        :return: a fake list of sites where the data is
        """
        logging.info("%s: Calling mock sitesByBlock", self.__class__.__name__)
        block = encodeUnicodeToBytesConditional(block, condition=PY3)
        # this algorithm gives same results in both python versions
        blockHash = int(hashlib.sha1(block).hexdigest()[:8], 16)
        if blockHash % 3 == 0:
            sites = ['T2_XX_SiteA']
        elif blockHash % 3 == 1:
            sites = ['T2_XX_SiteA', 'T2_XX_SiteB']
        else:
            sites = ['T2_XX_SiteA', 'T2_XX_SiteB', 'T2_XX_SiteC']
        return sites

    def __getattr__(self, item):
        """
        __getattr__ gets called in case lookup of the actual method fails. We use this to return data based on
        a lookup table

        :param item: The method name the user is trying to call
        :return: The generic lookup function
        """

        def genericLookup(*args, **kwargs):
            """
            This function returns the mocked DBS data

            :param args: positional arguments it was called with
            :param kwargs: named arguments it was called with
            :return: the dictionary that DBS would have returned
            """
            logging.info("%s: Calling mock genericLookup", self.__class__.__name__)
            if kwargs:
                for k in kwargs:
                    if isinstance(kwargs[k], (list, tuple)):
                        kwargs[k] = [encodeUnicodeToBytesConditional(x, condition=PY2) for x in kwargs[k]]
                    else:
                        kwargs[k] = encodeUnicodeToBytesConditional(kwargs[k], condition=PY2)
                signature = '%s:%s' % (item, sorted(kwargs.items()))
            else:
                signature = item

            try:
                return MOCK_DATA[signature]
            except KeyError:
                msg = "Rucio mock API failed to find key for signature: {}".format(signature)
                raise KeyError(msg)

        return genericLookup

    def getBlocksInContainer(self, container, scope='cms'):
        """
        Returns list of block names for given container
        """
        cname = self.__class__.__name__
        blockNames = [container + '#123', container + '#456']
        logging.info("%s getBlocksInContainer %s", cname, blockNames)
        return blockNames

    def listDataRules(self, name, **kwargs):
        """
        Emulate listDataRules Rucio API
        :return: list of dictionary records
        """
        cname = self.__class__.__name__
        logging.info("%s listDataRules name=%s kwargs=%s", cname, name, kwargs)
        return [{'rse_expression': 'T1_XX_ABC', 'id': 123, 'state': 'OK'}]

    def createContainer(self, name, scope='cms', **kwargs):
        """
        Create a CMS dataset (Rucio container) in a given scope.
        """
        cname = self.__class__.__name__
        logging.info("%s createContainer for name=%s scope=%s", cname, name, scope)
        return True

    def attachDIDs(self, rse, superDID, portion, scope='cms'):
        """
        Emulate attachDIDs Rucio API
        """
        cname = self.__class__.__name__
        logging.info("%s attachDID rse=%s, suportDID=%s, portion=%s, scope=%s", cname, rse, superDID, portion, scope)
        return True

    def createReplicationRule(self, portion, rseExpression, scope='cms'):
        """
        Emulate createReplicationRule Rucio API
        """
        cname = self.__class__.__name__
        logging.info("%s createReplicationRule portion=%s, rseExpression=%s", cname, portion, rseExpression)
        return [rseExpression]

    def updateRule(self, rid, opts):
        """
        Emulate updateRule(rid, opts) Rucio API
        """
        cname = self.__class__.__name__
        logging.info("%s updateRule rid=%s, opts=%s", cname, rid, opts)
        return True

    def getReplicaInfoForBlocks(self, **args):
        """
        Returns a mocked location for data.
        In the same format as the real `getReplicaInfoForBlocks` from the main module.
        """
        logging.info("%s: Calling mock getReplicaInfoForBlocks", self.__class__.__name__)
        result = []
        for blockName in args['block']:
            rses = self.sitesByBlock(block=blockName)
            result.append(dict(name=blockName, replica=rses))
        return result

    def getDataLockedAndAvailable(self, **kwargs):
        """
        Mock the method to discover where data is locked and available.
        Note that, by default, it will not return any Tape RSEs.
        :return: a unique list of RSEs
        """
        logging.info("%s: Calling mock getDataLockedAndAvailable", self.__class__.__name__)
        if 'name' not in kwargs:
            raise WMRucioException("A DID name must be provided to the getBlockLockedAndAvailable API")
        if self.isContainer(kwargs['name']):
            # then resolve it at container level and all its blocks
            return self.getContainerLockedAndAvailable(**kwargs)

        if 'grouping' in kwargs:
            # long strings seem not to be working, like ALL / DATASET. Make it short!
            kwargs['grouping'] = kwargs['grouping'][0]
        # TODO: either grouping or returnTape should change this response...
        returnTape = kwargs.pop("returnTape", False)

        rses = set()
        if kwargs['name'].split('#')[0] == PILEUP_DATASET:
            # Pileup is at a single site
            sites = ['T2_XX_SiteC']
            _BLOCK_LOCATIONS[kwargs['name']] = sites
        else:
            sites = self.sitesByBlock(block=kwargs['name'])
            _BLOCK_LOCATIONS[kwargs['name']] = sites
        rses.update(sites)
        return list(rses)

    def getPileupLockedAndAvailable(self, container, account, scope="cms"):
        """
        Mock method to resolve where the pileup container (and all its blocks)
        is locked and available.
        """
        logging.info("%s: calling mock getPileupLockedAndAvailable", self.__class__.__name__)
        result = dict()
        if not self.isContainer(container):
            raise WMRucioException("Pileup location needs to be resolved for a container DID type")

        kwargs = dict(name=container, account=account, scope=scope)

        try:
            DBS3Reader(PROD_DBS).checkDatasetPath(kwargs['name'])
            blocks = DBS3Reader(PROD_DBS).listFileBlocks(dataset=kwargs['name'])
            for block in blocks:
                result[block] = self.sitesByBlock(block)
        except DBSReaderError:
            logging.error("%s: Failed to fetch blocks from DBS", self.__class__.__name__)
        return result

    def getContainerLockedAndAvailable(self, **kwargs):
        """
        Mock the method to discover where container data is locked and available.
        Note that, by default, it will not return any Tape RSEs.
        :return: a unique list of RSEs
        """
        logging.info("%s: Calling mock getContainerLockedAndAvailable", self.__class__.__name__)
        if 'name' not in kwargs:
            raise WMRucioException("A DID name must be provided to the getContainerLockedAndAvailable API")
        kwargs.setdefault("scope", "cms")

        if kwargs['name'] == PILEUP_DATASET:
            return ['T2_XX_SiteA', 'T2_XX_SiteB', 'T2_XX_SiteC']
        try:
            DBS3Reader(PROD_DBS).checkDatasetPath(kwargs['name'])
            blocks = DBS3Reader(PROD_DBS).dbs.listBlocks(dataset=kwargs['name'])
            singleBlock = blocks[0]['block_name']
            return self.sitesByBlock(singleBlock)
        except DBSReaderError:
            return []

    def isContainer(self, didName, scope='cms'):
        """
        Mock check for whether a DID name corresponds to a container type or not,
        by simply relying on the naming convention
        :param didName: string with the DID name
        :param scope: string containing the Rucio scope (defaults to 'cms')
        :return: True if the DID is a container, else False
        """
        # TODO: figure use cases where we need to raise this exception
        if didName == "a bad DID name yet to be defined":
            msg = "Data identifier not found in MockRucio: {}".format(didName)
            raise WMRucioDIDNotFoundException(msg)
        return "#" not in didName



#!/usr/bin/env python
"""
_DeleteFileset_

MySQL implementation of DeleteCheck

"""
__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class DeleteCheck(DBFormatter):
    sql = """DELETE FROM wmbs_workflow WHERE id = :workflow AND
           NOT EXISTS (SELECT id FROM wmbs_subscription WHERE workflow = :workflow AND id != :subscription)"""

    def execute(self, workid = None, subid = None, conn = None, transaction = False):
        self.dbi.processData(self.sql, {'workflow': workid, 'subscription': subid},
                         conn = conn, transaction = transaction)
        return True #or raise



#!/usr/bin/python
"""
_ProdException_

General Exception class for Prod modules

"""

from future.utils import viewitems

try:
    import exceptions
except ImportError:
    import builtins as exceptions

import inspect
import logging


class ProdException(exceptions.Exception):
    """
    _ProdException_

    Exception class which works out details of where
    it was raised.

    """
    def __init__(self, message, errorNo=1000,**data):
        self.name = str(self.__class__.__name__)
        exceptions.Exception.__init__(self, self.name,
                                      message)

        #  //
        # // Init data dictionary with defaults
        #//
        self.data = {}
        self.data.setdefault("ClassName", None)
        self.data.setdefault("ModuleName", None)
        self.data.setdefault("MethodName", None)
        self.data.setdefault("ClassInstance", None)
        self.data.setdefault("FileName", None)
        self.data.setdefault("LineNumber", None)
        if errorNo==None:
            self.data.setdefault("ErrorNr",0)
        else:
            self.data.setdefault("ErrorNr",errorNo)

        self.message = message
        self.data.update(data)

        #  //
        # // Automatically determine the module name
        #//  if not set
        if self.data['ModuleName'] == None:
            frame = inspect.currentframe()
            lastframe = inspect.getouterframes(frame)[1][0]
            excepModule = inspect.getmodule(lastframe)
            if excepModule != None:
                modName = excepModule.__name__
                self.data['ModuleName'] = modName


        #  //
        # // Find out where the exception came from
        #//
        stack = inspect.stack(1)[1]
        self.data['FileName'] = stack[1]
        self.data['LineNumber'] = stack[2]
        self.data['MethodName'] = stack[3]

        #  //
        # // ClassName if ClassInstance is passed
        #//
        if self.data['ClassInstance'] != None:
            self.data['ClassName'] = \
              self.data['ClassInstance'].__class__.__name__

        logging.error(str(self))

    def __getitem__(self, key):
        """
        make exception look like a dictionary
        """
        return self.data[key]

    def __setitem__(self, key, value):
        """
        make exception look like a dictionary
        """
        self.data[key] = value

    def addInfo(self, **data):
        """
        _addInfo_

        Add key=value information pairs to an
        exception instance
        """
        for key, value in viewitems(data):
            self.data[key] = value
        return

    def xml(self):
        """create a xml string rep of this exception"""
        strg ="<Exception>\n"
        strg +="<Object>\n"
        strg += "%s\n" % self.name
        strg +="</Object>\n"
        strg +="<Message>\n"
        strg += self.message
        strg +="</Message>\n"
        strg +="<DataItems>\n"
        for key, value in viewitems(self.data):
            strg +="<DataItem>\n"
            strg += "<Key>\n"
            strg += str(key)
            strg += "</Key>\n"
            strg += "<Value>\n"
            strg += str(value)
            strg += "</Value>\n"
            strg +="</DataItem>\n"
        strg +="</DataItems>\n"
        strg +="</Exception>\n"
        logging.error(strg)
        return strg

    def __str__(self):
        """create a string rep of this exception"""
        strg = "%s\n" % self.name
        strg += "Message: %s\n" % self.message
        for key, value in viewitems(self.data):
            strg += "\t%s : %s\n" % (key, value, )
        return strg



from __future__ import (print_function, division)
from future.utils import viewitems

import json
import traceback

import cherrypy

import WMCore.ReqMgr.Service.RegExp as rx
from Utils.Utilities import strToBool
from WMCore.REST.Format import JSONFormat, PrettyJSONFormat
from WMCore.REST.Server import RESTEntity, restcall
from WMCore.REST.Tools import tools
from WMCore.REST.Validation import validate_str
from WMCore.REST.Error import MethodWithoutQueryString
from WMCore.ReqMgr.DataStructs.ReqMgrConfigDataCache import ReqMgrConfigDataCache
from WMCore.ReqMgr.DataStructs.RequestError import InvalidSpecParameterValue
from WMCore.ReqMgr.Utils.Validation import get_request_template_from_type
from WMCore.WMSpec.WMWorkload import WMWorkloadHelper


def format_algo_web_list(task_name, task_type, split_param, algo_config):
    fdict = {"taskName": task_name}
    fdict["taskType"] = task_type
    default_algo = split_param["algorithm"]
    algo_list = algo_config["algo_list_by_types"][task_type]
    param_list = []

    if default_algo in algo_list:
        new_param = {"algorithm": default_algo}
        for key, value in viewitems(split_param):
            if key in algo_config["algo_params"][default_algo]:
                new_param[key] = value
        param_list.append(new_param)
    elif default_algo == "":
        raise cherrypy.HTTPError(400, "Algorithm name is empty: %s" % split_param)
    else:
        param_list.append(split_param)

    # If task type is merge don't allow change the algorithm
    if fdict["taskType"] != "Merge":
        for algo in algo_list:
            if algo != default_algo:
                param = {"algorithm": algo}
                param.update(algo_config["algo_params"][algo])
                param_list.append(param)

    fdict["splitParamList"] = param_list
    return fdict


def create_web_splitting_format(split_info):
    web_form = []
    splitSettings = ReqMgrConfigDataCache.getConfig("EDITABLE_SPLITTING_PARAM_CONFIG")

    for sp in split_info:
        # skip Cleanup and LogCollect: don't allow change the param
        if sp["taskType"] not in ["Cleanup", "LogCollect"]:
            web_form.append(format_algo_web_list(sp["taskName"], sp["taskType"],
                                                 sp["splitParams"], splitSettings))
    return web_form


def create_updatable_splitting_format(split_info):
    """
    _create_updatable_splitting_format_

    Returns the workflow job splitting without parameters that
    cannot be updated in the POST call.
    """
    splitInfo = []
    splitSettings = ReqMgrConfigDataCache.getConfig("EDITABLE_SPLITTING_PARAM_CONFIG")

    for taskInfo in split_info:
        if taskInfo["taskType"] not in ["Cleanup", "LogCollect"]:
            splittingAlgo = taskInfo["splitAlgo"]
            submittedParams = taskInfo["splitParams"]
            splitParams = {}
            for param in submittedParams:
                validFlag, _ = _validate_split_param(splittingAlgo, param, splitSettings)
                if validFlag:
                    splitParams[param] = taskInfo["splitParams"][param]
            taskInfo["splitParams"] = splitParams
            splitInfo.append(taskInfo)
    return splitInfo


def _validate_split_param(split_algo, split_param, algo_config):
    """
    validate param for editing, also returns param type
    """

    valid_params = algo_config["algo_params"][split_algo]
    if split_param in valid_params:
        if isinstance(valid_params[split_param], bool):
            cast_type = bool
        else:
            cast_type = int
        return (True, cast_type)
    else:
        return (False, None)


def _assign_key_value(keyname, keyvalue, return_params, cast_type):
    if cast_type is None:
        return_params[keyname] = keyvalue
    elif cast_type == bool:
        try:
            return_params[keyname] = strToBool(keyvalue)
        except ValueError:
            msg = "%s expects a boolean value, you provided %s" % (keyname, keyvalue)
            raise cherrypy.HTTPError(400, msg)
    else:
        return_params[keyname] = cast_type(keyvalue)


class RequestSpec(RESTEntity):
    def validate(self, apiobj, method, api, param, safe):
        """
        Validate request input data.
        Has to be implemented, otherwise the service fails to start.
        If it's not implemented correctly (e.g. just pass), the arguments
        are not passed in the method at all.

        """
        validate_str("name", param, safe, rx.RX_REQUEST_NAME, optional=False)

    @restcall(formats=[('text/plain', PrettyJSONFormat()), ('application/json', JSONFormat())])
    @tools.expires(secs=-1)
    def get(self, name):
        """
        Spec templete API call.

        :arg str name: name to appear in the result message.
        :returns: row with response, here 1 item list with message.

        """
        result = get_request_template_from_type(name)
        return [result]


class WorkloadConfig(RESTEntity):
    def __init__(self, app, api, config, mount):
        # main CouchDB database where requests/workloads are stored
        RESTEntity.__init__(self, app, api, config, mount)
        self.reqdb_url = "%s/%s" % (config.couch_host, config.couch_reqmgr_db)

    def _validate_args(self, param, safe):
        # TODO: need proper validation but for now pass everything
        args_length = len(param.args)
        if args_length == 1:
            safe.kwargs["name"] = param.args[0]
            param.args.pop()
        else:
            raise MethodWithoutQueryString
        return

    def validate(self, apiobj, method, api, param, safe):
        """
        Validate request input data.
        Has to be implemented, otherwise the service fails to start.
        If it's not implemented correctly (e.g. just pass), the arguments
        are not passed in the method at all.

        """

        self._validate_args(param, safe)

    @restcall(formats=[('text/plain', PrettyJSONFormat()), ('application/json', JSONFormat())])
    @tools.expires(secs=-1)
    def get(self, name):
        """
        Workload config world API call.

        :arg str name: name to appear in the result message.
        :returns: row with response.

        """

        helper = WMWorkloadHelper()
        try:
            helper.loadSpecFromCouch(self.reqdb_url, name)
        except Exception:
            raise cherrypy.HTTPError(404, "Cannot find workload: %s" % name)

        return str(helper.data)


class WorkloadSplitting(RESTEntity):
    def __init__(self, app, api, config, mount):
        # main CouchDB database where requests/workloads are stored
        RESTEntity.__init__(self, app, api, config, mount)
        self.reqdb_url = "%s/%s" % (config.couch_host, config.couch_reqmgr_db)

    def _validate_get_args(self, param, safe):
        # TODO: need proper validation but for now pass everything
        args_length = len(param.args)
        if args_length == 1:
            safe.kwargs["name"] = param.args[0]
            param.args.pop()
        elif args_length == 2 and param.args[0] == "web_form":
            safe.kwargs["web_form"] = True
            safe.kwargs["name"] = param.args[1]
            param.args.pop()
            param.args.pop()
        elif args_length == 2 and param.args[0] == "update_only":
            safe.kwargs["update_only"] = True
            safe.kwargs["name"] = param.args[1]
            param.args.pop()
            param.args.pop()
        else:
            raise MethodWithoutQueryString
        return

    def validate(self, apiobj, method, api, param, safe):
        """
        Validate request input data.
        Has to be implemented, otherwise the service fails to start.
        If it's not implemented correctly (e.g. just pass), the arguments
        are not passed in the method at all.

        """
        try:
            if method == 'GET':
                self._validate_get_args(param, safe)

            if method == 'POST':
                args_length = len(param.args)
                if args_length == 1:
                    safe.kwargs["name"] = param.args[0]
                    param.args.pop()
        except InvalidSpecParameterValue as ex:
            raise ex
        except Exception as ex:
            msg = traceback.format_exc()
            cherrypy.log("Error: %s" % msg)
            if hasattr(ex, "message"):
                if hasattr(ex.message, '__call__'):
                    msg = ex.message()
                else:
                    msg = str(ex)
            else:
                msg = str(ex)
            raise InvalidSpecParameterValue(msg)

    @restcall(formats=[('text/plain', PrettyJSONFormat()), ('application/json', JSONFormat())])
    @tools.expires(secs=-1)
    def get(self, name, web_form=False, update_only=False):
        """
        getting job splitting algorithm.

        :arg str name: name to appear in the result message.
        :returns: row with response, here 1 item list with message.

        """

        helper = WMWorkloadHelper()
        try:
            helper.loadSpecFromCouch(self.reqdb_url, name)
        except Exception:
            raise cherrypy.HTTPError(404, "Cannot find workload: %s" % name)

        splittingDict = helper.listJobSplittingParametersByTask(performance=False)
        taskNames = sorted(splittingDict.keys())
        splitInfo = []
        for taskName in taskNames:
            splitInfo.append({"splitAlgo": splittingDict[taskName]["algorithm"],
                              "splitParams": splittingDict[taskName],
                              "taskType": splittingDict[taskName]["type"],
                              "taskName": taskName})
        if web_form:
            splitInfo = create_web_splitting_format(splitInfo)
        elif update_only:
            splitInfo = create_updatable_splitting_format(splitInfo)

        return splitInfo

    @restcall(formats=[('application/json', JSONFormat())])
    @tools.expires(secs=-1)
    def post(self, name):
        """
        Parse job splitting parameters sent from the splitting parameter update
        page.  Pull down the request and modify the new spec applying the
        updated splitting parameters.
        """

        data = cherrypy.request.body.read()
        splittingInfo = json.loads(data)
        cherrypy.log("Updating job splitting for '%s' with these args: %s" % (name, splittingInfo))

        helper = WMWorkloadHelper()
        try:
            helper.loadSpecFromCouch(self.reqdb_url, name)
        except Exception:
            raise cherrypy.HTTPError(404, "Cannot find workload for: %s" % name)

        splitSettings = ReqMgrConfigDataCache.getConfig("EDITABLE_SPLITTING_PARAM_CONFIG")
        for taskInfo in splittingInfo:
            splittingTask = taskInfo["taskName"]
            splittingAlgo = taskInfo["splitAlgo"]
            submittedParams = taskInfo["splitParams"]
            splitParams = {}
            for param in submittedParams:
                validFlag, castType = _validate_split_param(splittingAlgo, param, splitSettings)
                if validFlag:
                    _assign_key_value(param, submittedParams[param], splitParams, castType)
                else:
                    msg = "Parameter '%s' is not supported in the algorithm '%s'" % (param, splittingAlgo)
                    raise cherrypy.HTTPError(400, msg)

            helper.setJobSplittingParameters(splittingTask, splittingAlgo, splitParams, updateOnly=True)

        # Now persist all these changes in the workload
        url = "%s/%s" % (self.reqdb_url, name)
        result = helper.saveCouchUrl(url)

        return result



"""
This module contains a few basic WMCore-related utilitarian
MSPileup functions
"""

from Utils.CertTools import ckey, cert

from WMCore.Services.pycurl_manager import RequestHandler


def getPileupDocs(mspileupUrl, queryDict=None, method='GET'):
    """
    Fetch documents from MSPileup according to the query passed in using POST.

    :param mspileupUrl: string with the MSPileup url
    :param queryDict: dictionary with the MongoDB query to run
    :param method: string, defines which HTTP method to use
    :return: returns a list with all the pileup objects, or raises
        an exception in case of failure
    """
    queryDict = queryDict or {}
    mgr = RequestHandler()
    headers = {'Content-Type': 'application/json'}
    data = mgr.getdata(mspileupUrl, queryDict, headers, verb=method,
                       ckey=ckey(), cert=cert(), encode=True, decode=True)
    if data and data.get("result", []):
        if "error" in data["result"][0]:
            msg = f"Failed to retrieve MSPileup documents with query: {queryDict}"
            msg += f" and error message: {data}"
            raise RuntimeError(msg)
    return data["result"]



#!/usr/bin/env python
"""
_JobStatusByWorkflowAndSite_

Oracle implementation for loading a job by scheduler status
"""


from WMCore.BossAir.MySQL.JobStatusByWorkflowAndSite \
     import JobStatusByWorkflowAndSite as MySQLJobStatusByWorkflowAndSite

class JobStatusByWorkflowAndSite(MySQLJobStatusByWorkflowAndSite):
    pass



#!/usr/bin/env python
"""
Version of WMCore/Services/MSPileup intended to be used with mock or unittest.mock
"""


def getPileupDocs(mspileupUrl, queryDict=None, method='GET'):
    """
    Returns list of Pileup Documents.
    """
    print(f"Mocking MSPileup getPileupDocs: \
          url: {mspileupUrl}, query: {queryDict}, method: {method}")

    queryDict = queryDict or {}

    pdict = {
        # "pileupName": "/MinBias_TuneCP5_14TeV-pythia8/PhaseIITDRSpring19GS-106X_upgrade2023_realistic_v2_ext1-v1/GEN-SIM",
        "pileupName": "/GammaGammaToEE_Elastic_Pt15_8TeV-lpair/Summer12-START53_V7C-v1/GEN-SIM",
        "pileupType": "classic",
        "insertTime": 1680873642,
        "lastUpdateTime": 1706216047,
        "expectedRSEs": [
            "T2_XX_SiteA",
            "T2_XX_SiteB",
            "T2_XX_SiteC"
            ],
        "currentRSEs": [
            "T2_XX_SiteA",
            "T2_XX_SiteB",
            "T2_XX_SiteC"
            ],
        "fullReplicas": 1,
        "campaigns": [
            "Apr2023_Val"
            ],
        "containerFraction": 1.0,
        "replicationGrouping": "ALL",
        "activatedOn": 1706216047,
        "deactivatedOn": 1680873642,
        "active": True,
        "pileupSize": 1233099715874,
        "ruleIds": [
            "55e5a21aecb5445c8aa40581a7bf18d2",
            "67a3fa7252f54507ba1c45f271beb754"
            ],
        "customName": "",
        "transition": []}

    if 'filters' in queryDict.keys():
        return [{k: v for k, v in pdict.items() if k in queryDict['filters']}]

    return [pdict]



#!/usr/bin/env python
# I'm afraid *args, **kwargs magic is needed here
#-*- coding: ISO-8859-1 -*-
"""
Rest Model abstract implementation
"""
from builtins import str, bytes

from Utils.Utilities import decodeBytesToUnicode, encodeUnicodeToBytes
from Utils.PythonVersion import PY3

from functools import wraps
from WMCore.Lexicon import check
from WMCore.WebTools.WebAPI import WebAPI
from cherrypy import request, HTTPError
import time
import cherrypy
import traceback
from datetime import datetime

def restexpose(func):
    """
    Decorate a function to clearly mark it as an exposed REST method
    """
    setattr(func, 'restexposed', True)
    setattr(func, 'exposed', False)
    return func

class RESTModel(WebAPI):
    """
    Rest model class interface. Subclass this method and add methods to
    self.methods, *do not* modify/override the handler method unless you really
    know what you're doing.
    """
    def __init__(self, config):
        WebAPI.__init__(self, config)
        self.methods = {}
        self.daofactory = False
        # If not set expire data after 5 mins
        self.defaultExpires = config.default_expires

    def _classifyHTTPError(self, verb, args, kwargs):
        """
        There are a few cases where input will cause an error which needs
        appropriate classification:
            - request uses a globally unsupported verb (raise 501)
            - request is to a non-existant method (raise 404)
            - request is to a valid method, and uses a valid VERB but the
              VERB isn't supported by the method
        """

        method = args[0]

        he = False
        # Make sure the verb and method are well named
        try:
            check("^[A-Z]+$", verb)
            check("^[a-z][A-Za-z0-9]+$", method)
        except:
            he = HTTPError(400, 'bad VERB or method')

        # Checks whether verb is supported, if not return 501 error
        if verb not in self.methods:
            data =  "Unsupported verb: %s" % (verb)
            he = HTTPError(501, data)
        else:
            # We know how to deal with this VERB
            # Do we know the method?
            method_for_verb = method in self.methods[verb]

            if method_for_verb:
                return
            # Is the method supported for a VERB different to the the requests?
            unsupported_verb = False

            other_verbs = list(self.methods)
            other_verbs.remove(verb)

            for v in other_verbs:
                unsupported_verb = unsupported_verb | (method in self.methods[v])

            # Checks whether method exists but used wrong verb
            if unsupported_verb:
                data =  "Unsupported method for %s: %s" % (verb, method)
                he = HTTPError(405, data)

            # The method doesn't exist
            elif not method_for_verb and not he:
                data = "Method not found for the verb %s: %s" % (verb, method)
                he = HTTPError(404, data)

        if he:
            self.debug(he.args[1])
            self.debug(traceback.format_exc())
            raise he

    def handler(self, verb, args, kwargs):
        """
        Call the appropriate method from self.methods for a given VERB. args are
        URI path elements, the first (arg[0]) is the method name, other elements
        (arg[1:]) are positional arguments to the method. kwargs are query string
        parameters (e.g. method?thing1=abc&thing2=def). All args and kwargs are
        passed to the model method, this is so configuration for a given method
        can be identified.
        """

        verb = verb.upper()
        if not args:
            args = []
        if not kwargs:
            kwargs = {}
        self._classifyHTTPError(verb, args, kwargs)
        try:
            method = args[0]
            methodCall = self.methods[verb][method]['call']
            # if there is nothing to be processed about the parameter
            # it will just return original parameters
            params, kwargs = self._processParams(args[1:], kwargs)
            if not getattr(methodCall, 'restexposed', False):
                msg = 'Using an implicitly exposed method, '
                msg += 'please use _addDAO, _addMethod or @restexposed around %s' % method
                self.warning(msg)
            data = methodCall(*params, **kwargs)
        # If a type error is raised the data from the client is bad - 400 error
        except TypeError as e:
            error = e.__str__()
            self.debug(error)
            self.debug(traceback.format_exc())
            raise HTTPError(400, error)
        # Don't need to handle other exceptions here - that's done in RESTAPI
        if 'expires' in self.methods[verb][method]:
            return data, self.methods[verb][method]['expires']
        else:
            return data, self.defaultExpires

    def _addDAO(self, verb, methodKey, daoStr, args=[], validation=[],
               version=1, daoFactory = None, expires = None,
               secured=False, security_params={}):
        """
        add dao in self.methods and wrap it with _sanitise_input. Assumes that a
        DAOFactory instance is available from self.
        """
        def function(*args, **kwargs):
            """
            Decorator to process the input by a DAO
            """
            # store the method name
            method = methodKey
            input_data = self._sanitise_input(args, kwargs, method)

            # store the dao
            if daoFactory:
                dao = daoFactory(classname=daoStr)
            else:
                assert self.daofactory, "Cannot _addDAO - do not have a daofactory instance"
                # if the code gets to here the dafactory is a daofactory instead of False
                # disable-msg=E1102
                dao = self.daofactory(classname=daoStr)
                # enable-msg=E1102
            # execute the requested input, now a list of keywords
            return dao.execute(**input_data)

        if expires == None:
            expires = self.defaultExpires
        self._addMethod(verb, methodKey, function, args, validation,
                       version, expires, secured, security_params)

    def _addMethod(self, verb, methodKey, function, args=[],
                  validation=[], version=1, expires = None,
                  secured=False, security_params={}, dump_request_info=False):
        """
        Add a method handler to self.methods self.methods, decorate it such that it
        receives sanitised input and is marked as 'restexposed'.
        """

        if verb not in self.methods:
            self.methods[verb] = {}
        @wraps(function)
        def wrapper(*input_args, **input_kwargs):
            if dump_request_info:
                request = cherrypy.request
                headers = dict(request.headers)
                keysToRemove = ['cookie', 'cms-authn-hmac', 'cms-authz-admin']
                for key in headers:
                    if key.lower().startswith('ssl-') or \
                        key.lower() in keysToRemove:
                        del headers[key]
                msg = 'REQUEST {} {} {} {} {} [{}] [{}]'.format(\
                        datetime.now().strftime('[%d/%b/%Y %H:%M:%S]'),
                        request.remote.ip,
                        request.remote.port,
                        request.method,
                        request.path_info,
                        headers.get('Cms-Authn-Dn', ''),
                        request.params
                        )
                cherrypy.log.access_log.info(msg)
            if secured:
                # set up security
                security = cherrypy.tools.secmodv2
                # security_params should be a dict like:
                #{'role':[], 'group':[], 'site':[], 'authzfunc':None}
                security.callable(
                                role=security_params.get('role', []),
                                group=security_params.get('group', []),
                                site=security_params.get('site', []),
                                authzfunc=security_params.get('authzfunc', None)
                                )
            if len(args) != 0:
                # store the method name
                method = methodKey
                input_data = self._sanitise_input(input_args, input_kwargs, method)
                # store the function
                func = function
                # execute the requested input, now a list of keywords
                return func(**input_data)
            else:
                # If the method isn't meant to have arguments the wrapper isn't needed
                # so we need to raise a 400 error
                if (len(input_args) + len(input_kwargs)):
                    raise HTTPError(400, 'Invalid input: Arguments added where none allowed')
                return function()

        funcRef = wrapper

        setattr(funcRef, 'restexposed', True)

        if expires == None:
            expires = self.defaultExpires
        self.methods[verb][methodKey] = {'args': args,
                                         'call': funcRef,
                                         'validation': validation,
                                         'version': version,
                                         'expires': expires}

    def _sanitise_input(self, input_args=[], input_kwargs={}, method = None):
        """
        Pull out the necessary input from kwargs (by name) and, failing that,
        pulls out the number required args from args, which assumes the
        arguments are positional.

        _sanitise_input is called automatically if you use the _addMethod/_addDAO
        convenience functions. If you add your method to the methods dictionary
        by hand you should call _sanitise_input explicitly.

        In all but the most basic cases you'll likely want to over-ride this, or
        at least treat its outcome with deep suspicion.

        TODO: Would be nice to loose the method argument and derive it in this method.

        Returns a dictionary of validated, sanitised input data.
        """
        verb = request.method.upper()

        if len(input_args):
            input_args = list(input_args)
        if (len(input_args) + len(input_kwargs)) > len(self.methods[verb][method]['args']):
            self.debug('%s to %s expects %s argument(s), got %s' % (verb, method, len(self.methods[verb][method]['args']), (len(input_args) + len(input_kwargs))))
            raise HTTPError(400, 'Invalid input: Input arguments failed sanitation.')
        input_data = {}

        # VK, we must read input kwargs/args as string types
        # rather then unicode one. This is important for cx_Oracle
        # driver which will place parameters into binded queries
        # due to mixmatch (string vs unicode) between python and Oracle
        # we must pass string parameters.
        for a in self.methods[verb][method]['args']:
            if a in input_kwargs:
                v = input_kwargs[a]
                input_data[a] = decodeBytesToUnicode(v) if PY3 else encodeUnicodeToBytes(v)
                input_kwargs.pop(a)
            else:
                if len(input_args):
                    v = input_args.pop(0)
                    input_data[a] = decodeBytesToUnicode(v) if PY3 else encodeUnicodeToBytes(v)
        if input_kwargs:
            raise HTTPError(400, 'Invalid input: Input arguments failed sanitation.')
        self.debug('%s raw data: %s' % (method, {'args': input_args, 'kwargs': input_kwargs}))
        self.debug('%s sanitised input_data: %s' % (method, input_data))
        return self._validate_input(input_data, verb, method)


    def _validate_input(self, input_data, verb, method):
        """
        Apply some checks to the input data. Run all the validation funstions
        for the given method. Validation functions should raise exceptions if
        the data doesn't pass the validation (assert's are your friend!). These
        exceptions are caught here and converted into 400 HTTPErrors.
        """
        validators = self.methods[verb][method].get('validation', [])
        if len(validators) == 0:
            # Do nothing
            return input_data
        result = {}
        # Run the validation functions, these should raise exceptions. If the
        # exception is an HTTPError re-raise it, else raise a 400 HTTPError (bad input).
        for fnc in validators:
            try:
                filteredInput = fnc(input_data)
            except HTTPError as he:
                self.debug(he)
                raise he
            except Exception as e:
                self.debug(e)
                raise HTTPError(400, 'Invalid input: Input data failed validation.')
            result.update(filteredInput)
        return result

    def _processParams(self, args, kwargs):
        """
        If the args and kwargs needs to be pre-processed (encoded, decoded) according
        to the http header values (i.e. content-type) or convert request.body to parameters,
        overwrite this function in child class

        WARNING: use this with caution, args are list of arguements and kwargs a dict of
        argument and value.

        Overwritten function should have the same return signature [], {}
        """
        return args, kwargs



#!/usr/bin/env python3
"""
Script to retrieve the request status and evaluate the workflow
completion, so the ration of input lumis vs output lumis.
"""
import argparse
import json
import os
import pwd
import sys
import ssl
from urllib.request import HTTPSHandler, build_opener
from urllib.parse import urlencode
from urllib.error import HTTPError, URLError
from http.client import HTTPSConnection

# ID for the User-Agent
CLIENT_ID = 'workflowCompletion::python/%s.%s' % sys.version_info[:2]


class HTTPSClientAuthHandler(HTTPSHandler):
    """
    Basic HTTPS class
    """

    def __init__(self, key, cert):
        HTTPSHandler.__init__(self)
        # Create a default ssl context manager to carry the credentials.
        # It also loads the default CA certificates
        self.sslContext = ssl.create_default_context()
        self.sslContext.load_cert_chain(cert, keyfile=key)
        # Also load the default CA certificates (apparently not needed)
        # self.sslContext.load_verify_locations(None, '/etc/grid-security/certificates/')

    def https_open(self, req):
        # Rather than pass in a reference to a connection class, we pass in
        # a reference to a function which, for all intents and purposes,
        # will behave as a constructor
        return self.do_open(self.getConnection, req)

    def getConnection(self, host, timeout=290):
        return HTTPSConnection(host, context=self.sslContext)


def getX509():
    "Helper function to get x509 from env or tmp file"
    certFile = os.environ.get('X509_USER_CERT', '')
    keyFile = os.environ.get('X509_USER_KEY', '')
    if certFile and keyFile:
        return certFile, keyFile

    proxy = os.environ.get('X509_USER_PROXY', '')
    if not proxy:
        proxy = '/tmp/x509up_u%s' % pwd.getpwuid(os.getuid()).pw_uid
        if not os.path.isfile(proxy):
            return '', ''
    return proxy, proxy


def getContent(url, params=None):
    certFile, keyFile = getX509()
    client = '%s (%s)' % (CLIENT_ID, os.environ.get('USER', ''))
    handler = HTTPSClientAuthHandler(keyFile, certFile)
    opener = build_opener(handler)
    opener.addheaders = [("User-Agent", client),
                         ("Accept", "application/json")]

    try:
        response = opener.open(url, params)
        output = response.read()
    except HTTPError as e:
        print(f"The server couldn't fulfill the request at {url}")
        print(f"Error: {e}")
        output = '{}'
        # sys.exit(1)
    except URLError as e:
        print(f'Failed to reach server at {url}')
        print(f'Reason: {e.reason}')
        sys.exit(2)
    return output


def handleReqMgr(reqName, reqmgrUrl):
    """
    Return the list of output datasets
    """
    urn = reqmgrUrl + "/reqmgr2/data/request/" + reqName
    reqmgrOut = json.loads(getContent(urn))['result'][0][reqName]

    if reqmgrOut['RequestStatus'] in ['assignment-approved', 'assigned', 'staging', 'staged']:
        print(f"Workflow {reqName} in status: {reqmgrOut['RequestStatus']} , skipping!\n")
        return None, None

    if 'InputDataset' in reqmgrOut:
        inputData = reqmgrOut['InputDataset']
    elif 'Task1' in reqmgrOut and 'InputDataset' in reqmgrOut['Task1']:
        inputData = reqmgrOut['Task1']['InputDataset']
    elif 'Step1' in reqmgrOut and 'InputDataset' in reqmgrOut['Step1']:
        inputData = reqmgrOut['Step1']['InputDataset']
    else:
        inputData = None

    print(f"==> {reqName}\t(status: {reqmgrOut['RequestStatus']})")
    print(f"InputDataset:\n    {inputData} (total input lumis: {reqmgrOut['TotalInputLumis']})")
    return reqmgrOut['TotalInputLumis'], reqmgrOut['OutputDatasets']


def handleDBS(reqmgrOutDsets, cmswebUrl):
    """
    Get total number of lumi sections in each dataset
    """
    if 'testbed' in cmswebUrl:
        dbsUrl = cmswebUrl + "/dbs/int/global/DBSReader"
    else:
        dbsUrl = cmswebUrl + "/dbs/prod/global/DBSReader"

    dbsOutput = {}
    for dataset in reqmgrOutDsets:
        fullUrl = dbsUrl + "/filesummaries?" + urlencode({'dataset': dataset})
        data = json.loads(getContent(fullUrl))
        if data:
            dbsOutput[dataset] = data[0]['num_lumi']
        else:
            dbsOutput[dataset] = 0

    return dbsOutput


def main():
    """
    Requirements: you need to have your proxy and proper x509 environment
     variables set.

    Receive a workflow name in order to fetch the following information:
     - from couchdb: gets the output datasets and spec file
     - from phedex: gets dataset, block, files information
     - from dbs: gets dataset, block and files information
    """
    parser = argparse.ArgumentParser(description="Validate workflow input, output and config")
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-w', '--workflow', help='A single workflow name')
    group.add_argument('-i', '--inputFile', help='Plain text file containing request names (one per line)')
    parser.add_argument('-c', '--cms', help='CMSWEB url to talk to DBS/PhEDEx. E.g: cmsweb-testbed.cern.ch')
    parser.add_argument('-r', '--reqmgr', help='Request Manager URL. Example: cmsweb-testbed.cern.ch')
    args = parser.parse_args()

    if args.workflow:
        listRequests = [args.workflow]
    elif args.inputFile:
        with open(args.inputFile, 'r', encoding="utf8") as f:
            listRequests = [req.rstrip('\n') for req in f.readlines()]
    else:
        parser.error("You must provide either a workflow name or an input file name.")
        sys.exit(3)

    cmswebUrl = "https://" + args.cms if args.cms else "https://cmsweb-prod.cern.ch"
    reqmgrUrl = "https://" + args.reqmgr if args.reqmgr else "https://cmsweb.cern.ch"

    for reqName in listRequests:
        ### Retrieve and process ReqMgr information
        inputLumis, reqmgrOutDsets = handleReqMgr(reqName, reqmgrUrl)
        if reqmgrOutDsets is None:
            continue

        ### Retrieve and process DBS information
        dbsInfo = handleDBS(reqmgrOutDsets, cmswebUrl)
        print("OutputDatasets:")
        for dset in reqmgrOutDsets:
            print(f"    {dset} (lumis: {dbsInfo[dset]}, lumi completion: {dbsInfo[dset] / inputLumis})")
        print("")

    sys.exit(0)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
"""
    WorkQueue tests
"""
from __future__ import absolute_import, print_function
from builtins import range, object

import os
import shutil
import tempfile

from WMCore.WMSpec.StdSpecs.ReReco  import ReRecoWorkloadFactory
from .Samples.BasicProductionWorkload \
    import getProdArgs, createWorkload as BasicProductionWorkload
from .Samples.BasicProcessingWorkload \
    import createWorkload as BasicProcessingWorkload
from WMCore.Cache.WMConfigCache import ConfigCache


mcArgs = getProdArgs()

class WMSpecGenerator(object):

    def __init__(self, dirLocation=None):
        if not dirLocation:
            dirLocation = tempfile.mkdtemp()
        if not os.path.exists(dirLocation):
            os.makedirs(dirLocation)
        self.dir = dirLocation

    def _save(self, spec):
        specFile =  os.path.join(self.dir, spec.name() + ".spec")
        # Basic production Spec
        spec.setSpecUrl(specFile)
        # save create pickled file
        spec.save(spec.specUrl())

        return specFile

    def _selectReturnType(self, spec, returnType, splitter = None):
        if splitter != None:
            spec.setStartPolicy(splitter)
        if returnType == "file":
            specUrl = self._save(spec)
            return specUrl
        elif returnType == "spec":
            return spec
        else:
            specUrl = self._save(spec)
            return spec, specUrl

    def createProductionSpec(self, specName, returnType="spec", splitter = None):
        spec = BasicProductionWorkload(specName)
        return self._selectReturnType(spec, returnType, splitter)

    def createProcessingSpec(self, specName, returnType="spec", splitter = None):
        #spec = BasicProcessingWorkload(specName)
        spec = BasicProcessingWorkload(specName)
        return self._selectReturnType(spec, returnType, splitter)

    def createReRecoSpec(self, specName, returnType="spec", splitter = None,
                         assignKwargs={}, **additionalArgs):
        # update args, then reset them
        args = ReRecoWorkloadFactory.getTestArguments()
        args.update(additionalArgs)
        args["ConfigCacheID"] = createConfig(args["CouchDBName"])
        factory = ReRecoWorkloadFactory()
        spec = factory.factoryWorkloadConstruction(specName, args)

        if assignKwargs:
            args = ReRecoWorkloadFactory.getAssignTestArguments()
            args.update(assignKwargs)
            spec.updateArguments(args)

        return self._selectReturnType(spec, returnType, splitter)

    def createRandomProductionSpecs(self, size=10):
        specNameBase = "FakeSpec"
        specs = {}
        for i in range(size):
            specName = specNameBase + "_%s" % i
            _, specUrl = self.createProductionSpec(specName)
            specs[specName] = specUrl
        return specs

    def removeSpecs(self):
        #TODO: needs a smarter way to clean up the spec files.
        # avoid accidently deleting other codes
        shutil.rmtree(self.dir)

def createConfig(couchDBName):
    """
    _createConfig_

    Create a config of some sort that we can load out of ConfigCache
    """

    PSetTweak = {'process': {'outputModules_': ['RECOoutput', 'ALCARECOoutput'],
                             'RECOoutput': {'dataset': {'dataTier': 'RECO',
                                                         'filterName': 'Filter'}},
                             'ALCARECOoutput': {'dataset': {'dataTier': 'ALCARECO',
                                                            'filterName': 'AlcaFilter'}}}}

    configCache = ConfigCache(os.environ["COUCHURL"], couchDBName = couchDBName)
    configCache.createUserGroup(groupname = "testGroup", username = 'testOps')
    configCache.setPSetTweaks(PSetTweak = PSetTweak)
    configCache.save()

    return configCache.getCouchID()




from __future__ import (division, print_function)
from builtins import object

class MockLogDB(object):
    """
    Version of Services/PhEDEx intended to be used with mock or unittest.mock
    """

    def __init__(self, *args, **kwargs):
        print("Using MockLogDB")

    def post(self, request=None, msg="", mtype="comment"):
        """Post new entry into LogDB for given request"""

        return 'OK'

    def get(self, request=None, mtype=None):
        """Retrieve all entries from LogDB for given request"""
        # TODO Need to add mock data
        return []

    def get_all_requests(self):
        """Retrieve all entries from LogDB for given request"""
        # TODO Need to add mock data
        return []

    def delete(self, request=None, mtype=None, this_thread=False):
        """
        Delete entry in LogDB for given request
        if mtype == None - delete all the log for that request
        mtype != None - only delete specified mtype
        """

        return "OK"

    def wmstats_down_components_report(self, thread_list):
        # TODO Need to add mock data
        return {}


#!/usr/bin/env python
"""
_GetFWJRTaskName_

MySQL implementation of Jobs.GetFWJRTaskName
"""

from WMCore.Database.DBFormatter import DBFormatter

class GetFWJRTaskName(DBFormatter):
    """
    _GetFWJRTaskName_

    Retrieve the taskName and the fwjr_path to heal corrupted FWJRs
    seen by the JobAccountant. See #5421.
    """
    sql = """
          SELECT wj.fwjr_path, ww.task FROM wmbs_workflow ww
            INNER JOIN wmbs_subscription ws ON ws.workflow = ww.id
            INNER JOIN wmbs_jobgroup wjg ON wjg.subscription = ws.id
            INNER JOIN wmbs_job wj ON wj.jobgroup = wjg.id
            WHERE wj.id = :jobId"""

    def format(self, results):
        """
        _format_

        """
        result = DBFormatter.format(self, results)

        return {"fwjr_path": result[0][0], "taskName": result[0][1]}

    def execute(self, jobId, conn = None, transaction = False):
        result = self.dbi.processData(self.sql, {"jobId": jobId}, conn = conn,
                                      transaction = transaction)

        return self.format(result)
